{-----------------------------------------------------------------------------
The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in compliance
with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/MPL-1.1.html

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either expressed or implied. See the License for
the specific language governing rights and limitations under the License.

The Original Code is: JvCCLists.PAS, released on 2000-12-01.

The Initial Developer of the Original Code is Michael Rynn.
Portions created by Michael Rynn are Copyright (C) 2000  Michael Rynn.
All Rights Reserved.

Contributor(s): ______________________________________.

Last Modified: 2000-12-01

You may retrieve the latest version of this file at the Project JEDI home page,
located at http://www.delphi-jedi.org

-----------------------------------------------------------------------------}
{$A+,B-,C+,D+,E-,F-,G+,H+,I+,J+,K-,L+,M-,N+,O+,P+,Q-,R-,S-,T-,U-,V+,W-,X+,Y+,Z1}

{$I JCL.INC}
unit JvCCLists;

interface
Uses JvCCRecrd;

type
  PCCSLink = ^TJvCCSLink;
  TJvCCSLink = record
    FNext : PCCSLink;
  end;

  PCCDLink = ^TJvCCDLink;
  TJvCCDLink = record
    FNext : PCCDLink;
    FPrior : PCCDLink;
  end;

  TJvCCLinkCursor = record
    CASE integer of
    0 : ( FSPrior : PCCSLink;
          FSLink : PCCSLink );
    1 : ( FDLink : PCCDLink );
  end;

  TJvCCLinkedList = class(TJvCCBase)
  public
    procedure SetCount(inCount : TJvCCSize); override;
    procedure AppendLast(const buf); override;
    // link specific
    procedure DeleteFirst; virtual; abstract;
    procedure DeleteLast; virtual; abstract;
  end;

  TJvCCSList = class(TJvCCLinkedList)
  protected
    FStartLink : PCCSLink;
    FEndLink : PCCSLink;
    
    function NewLink(const data): PCCSLink;
    procedure FreeLink(plink : PCCSLink);
    procedure FreeFromChain(prev, linktofree : PCCSLink);

  public
    constructor CreateMgr(reclass : TJvCCRecordMgrClass); override;
    procedure Clear; override;
    procedure FreeItem(var pitem : TJvCCItemPtr);  override;
    function FirstItem : TJvCCItemPtr; override;
    function LastItem : TJvCCItemPtr; override;
    function NextItem(pitem : TJvCCItemPtr) : TJvCCItemPtr; override;
    function PriorItem(pitem : TJvCCItemPtr) : TJvCCItemPtr; override;
    function AppendItem(const buf): TJvCCItemPtr; override;
    function DeleteItem(const buf): boolean; override;
    procedure AppendFirst(const buf);override;

    procedure DeleteFirst; override;
    procedure DeleteLast; override;

   end;

  TJvCCSListIterator = class(TObject)
  protected
    FSList : TJvCCSList;
    FCurrLink : PCCSLink;
    procedure NotifyDeleted(invalidlink,replacelink:PCCSLink);
  public
    constructor Create(slist:TJvCCSList);
    function First : boolean;
    function Next : boolean;
    function Data : Pointer;
  end;

  TJvCCDList = class(TJvCCLinkedList)
  protected
    FStartLink : PCCDLink;
    FEndLink : PCCDLink;
  public
    function NewLink(const data): PCCDLink;
    procedure FreeLink(plink : PCCDLink);
    procedure FreeFromChain(linktofree : PCCDLink);
  public
    constructor CreateMgr(reclass : TJvCCRecordMgrClass); override;
    procedure Clear; override;
    procedure FreeItem(var pitem : TJvCCItemPtr);  override;
    function FirstItem : TJvCCItemPtr; override;
    function LastItem : TJvCCItemPtr; override;
    procedure DeleteCCItemPtr(pitem : TJvCCItemPtr);
    function NextItem(pitem : TJvCCItemPtr) : TJvCCItemPtr; override;
    function PriorItem(pitem : TJvCCItemPtr) : TJvCCItemPtr; override;
    function AppendItem(const buf) : TJvCCItemPtr; override;
    function DeleteItem(const buf): boolean; override;
    procedure AppendFirst(const buf);override;
    procedure DeleteFirst; override;
    procedure DeleteLast; override;
  end;

  TJvCCDListIterator = class(TObject)
  protected
    FDList : TJvCCDList;
    FCurrLink : PCCDLink;
    procedure NotifyDeleted(invalidlink,replacelink:PCCDLink);
  public
    constructor Create(dlist:TJvCCDList);
    function First : boolean;
    function Last : boolean;
    function Next : boolean;
    function Prior : boolean;
    function Data : Pointer;
  end;

implementation

procedure TJvCCLinkedList.SetCount(inCount : TJvCCSize);
begin
  {no can set count externally}
end;

constructor TJvCCSList.CreateMgr(reclass : TJvCCRecordMgrClass);
begin
  inherited CreateMgr(reclass);
  FCount := 0;
  FStartLink := nil;
  FEndLink := nil;
end;

function TJvCCSList.NewLink(const data): PCCSLink;
var
  linksize : integer;
  pdata : PCCSLink;
begin
  linksize := sizeof(TJvCCSLink) + FRecMgr.ItemSize;
  GetMem(Result,linksize);
  pdata := Result;
  pdata^.FNext := nil;
  Inc(pdata,1);
  if FRecMgr.InitItems then
    FRecMgr.ConstructItem(pdata^);
  FRecMgr.AssignItem(pdata^,data);
end;

procedure TJvCCSList.Clear;
var
  plink : PCCSLink;
  nlink : PCCSLink;
begin
  plink := FStartLink;
  while (plink <> nil) do
  begin
    nlink := plink^.FNext;
    FreeFromChain(nil,plink);
    FreeLink(plink);
    plink := nlink;
  end;
end;

function TJvCCSList.AppendItem(const buf) : TJvCCItemPtr;
var
  plink : PCCSLink;
begin
  plink := NewLink(buf);
  if (FEndLink=nil) then
  begin
    FEndLink := plink;
    FStartLink := plink;
  end else
  begin
    FEndLink^.FNext := plink;
    FEndLink := plink;
  end;
  Inc(FCount);
  Inc(plink);
  Result := plink;
end;

procedure TJvCCSList.FreeLink(plink : PCCSLink);
var
  linksize : integer;
  dlink : PCCSLink;{data}
begin
  dlink := plink;
  Inc(dlink,1);
  if FRecMgr.InitItems then
    FRecMgr.DestructItem(dlink^);
  linksize := sizeof(TJvCCSLink) + FRecMgr.ItemSize;
  FreeMem(plink,linksize);
end;

procedure TJvCCSList.FreeFromChain(prev, linktofree : PCCSLink);
var
  nlink : PCCSLink;{next}
begin
  nlink := linktofree^.FNext;
  if (prev <> nil) then
    prev^.FNext := nlink;
  if (linktofree = FEndLink) then
    FEndLink := prev;
  if (linktofree = FStartLink) then
    FStartLink := nlink;
  linktofree^.FNext := nil;
  Dec(FCount);
end;

function TJvCCSList.DeleteItem(const buf) : boolean;
var
  plink : PCCSLink;{previous}
  slink : PCCSLink;{test link}
  nlink : PCCSLink;{next}
  dlink : PCCSLink;{data}
begin
  slink := FStartLink;
  plink := nil;
  while (slink <> nil) do
  begin
    nlink := slink^.FNext;
    dlink := slink;
    Inc(dlink,1);
    if FRecMgr.CompareItems(dlink^,buf)=0 then
    begin
      FreeFromChain(plink,slink);
      FreeLink(slink);
      Result := TRUE;
      exit;
    end;
    plink := slink;
    slink := nlink;
  end;
  Result := FALSE;
end;

procedure TJvCCSList.FreeItem(var pitem : TJvCCItemPtr);
var
  dlink : PCCSLink;{link to free}
  plink : PCCSLink;{prior link}
  nlink : PCCSLink;{next}
begin
  dlink := pitem;
  if (dlink = nil) then
    exit;
  Dec(dlink,1);
  nlink := FStartLink;
  plink := nil;
  while (nlink <> nil) do
  begin
    if (nlink=dlink) then
    begin
      FreeFromChain(plink,nlink);
      FreeLink(dlink);
      pitem := nil;
      exit;
    end;
    plink := nlink;
    nlink := nlink^.FNext;
  end;
end;


procedure TJvCCSList.AppendFirst(const buf);
var
  plink : PCCSLink;
begin
  plink := NewLink(buf);
  if (FStartLink=nil) then
  begin
    FEndLink := plink;
    FStartLink := plink;
  end else
  begin
    plink^.FNext := FStartLink;
    FStartLink := plink;
  end;
  Inc(FCount);
end;

procedure TJvCCLinkedList.AppendLast(const buf);
begin
  AppendItem(buf);
end;

procedure TJvCCSList.DeleteFirst;
var
  plink : PCCSLink;
begin
  plink := FStartLink;
  if (plink <> nil) then
  begin
    FreeFromChain(nil,plink);
    FreeLink(plink);
  end;
end;

procedure TJvCCSList.DeleteLast;
var
  slink : PCCSLink;
  plink : PCCSLink;
begin
  if FEndLink<>nil then
  begin
    plink := nil;
    slink := FStartLink;
    while (slink <> FEndLink) do
    begin
      plink := slink;
      slink := slink^.FNext;
    end;
    FreeFromChain(plink,slink);
    FreeLink(slink);
  end;
end;

function TJvCCSList.FirstItem : TJvCCItemPtr;
var
  plink : PCCSLink;
begin
  plink := FStartLink;
  if (plink <> nil) then
    Inc(plink,1);
  Result := plink;
end;

function TJvCCSList.LastItem : TJvCCItemPtr;
var
  plink : PCCSLink;
begin
  plink := FEndLink;
  if (plink <> nil) then
    Inc(plink,1);
  Result := plink;
end;


function TJvCCSList.NextItem(pitem : TJvCCItemPtr) : TJvCCItemPtr;
var {buf is assumed to be initialised}
  plink : PCCSLink;
begin
  plink := pitem;
  if (plink <> nil) then
  begin
    Dec(plink,1);
    plink := plink^.FNext;
    if (plink <> nil) then
      Inc(plink,1);
  end;
  Result := plink;
end;

function TJvCCSList.PriorItem(pitem : TJvCCItemPtr) : TJvCCItemPtr;
var {buf is assumed to be initialised}
  plink : PCCSLink;
  nlink : PCCSLink;
  rlink : PCCSLink;
begin
  rlink := pitem;
  if (rlink <> nil) then
  begin
    Dec(rlink,1);
    plink := nil;
    nlink := FStartLink;
    while (nlink <> nil) do
    begin
      if (nlink=rlink) then
      begin
        if (plink <> nil) then
        begin
          Inc(plink,1);
          Result := plink;
          exit;
        end;
      end;
      plink := nlink;
      nlink := nlink^.FNext;
    end;
  end;
  Result := nil;
end;


constructor TJvCCSListIterator.Create(slist:TJvCCSList);
begin
  inherited Create;
  FSList := slist;
end;

function TJvCCSListIterator.First : boolean;
begin
  FCurrLink := FSList.FStartLink;
  Result  := (FCurrLink <> nil);
end;

function TJvCCSListIterator.Next: boolean;
begin
  FCurrLink := FCurrLink.FNext;
  Result  := (FCurrLink <> nil);
end;

function TJvCCSListIterator.Data : Pointer;
var
  rp:PCCSLink;
begin
  rp := FCurrLink;
  if (rp <> nil) then
    Inc(rp,1);
  Result := rp;
end;

procedure TJvCCSListIterator.NotifyDeleted(invalidlink,replacelink:PCCSLink);
begin
  if (invalidLink = FCurrLink) then
    FCurrLink := replacelink;
end;

function TJvCCDList.NewLink(const data): PCCDLink;
var
  linksize : integer;
  pdata : PCCDLink;
begin
  linksize := sizeof(TJvCCDLink) + FRecMgr.ItemSize;
  GetMem(Result,linksize);
  pdata := Result;
  pdata^.FNext := nil;
  pdata^.FPrior := nil;
  Inc(pdata,1);
  if FRecMgr.InitItems then
    FRecMgr.ConstructItem(pdata^);
  FRecMgr.AssignItem(pdata^,data);
end;

procedure TJvCCDList.FreeLink(plink : PCCDLink);
var
  linksize : integer;
  dlink : PCCDLink;{data}
begin
  dlink := plink;
  Inc(dlink,1);
  if FRecMgr.InitItems then
    FRecMgr.DestructItem(dlink^);
  linksize := sizeof(TJvCCDLink) + FRecMgr.ItemSize;
  FreeMem(plink,linksize);
end;

procedure TJvCCDList.FreeFromChain(linktofree : PCCDLink);
var
  nlink : PCCDLink;{next}
  plink : PCCDLink;{prior}
begin
  nlink := linktofree^.FNext;
  plink := linktofree^.FPrior;
  if (plink <> nil) then
    plink^.FNext := nlink
  else
  begin
    FStartLink := nlink;
  end;

  if (nlink <> nil) then
    nlink^.FPrior := plink
  else
  begin
    FEndLink := plink;
  end;
  linktofree^.FNext := nil;
  linktofree^.FPrior := nil;
  Dec(FCount);
end;


constructor TJvCCDList.CreateMgr(reclass : TJvCCRecordMgrClass);
begin
  inherited CreateMgr(reclass);
  FStartLink := nil;
  FEndLink := nil;
end;

procedure TJvCCDList.Clear;
var
  plink : PCCDLink;
  nlink : PCCDLink;
begin
  plink := FStartLink;
  while (plink <> nil) do
  begin
    nlink := plink^.FNext;
    FreeFromChain(plink);
    FreeLink(plink);
    plink := nlink;
  end;
end;

procedure TJvCCDList.AppendFirst(const buf);
var
  plink : PCCDLink;
begin
  plink := NewLink(buf);
  if (FStartLink=nil) then
  begin
    FEndLink := plink;
    FStartLink := plink;
  end else
  begin
    FStartLink^.FPrior := plink;
    plink^.FNext := FStartLink;
    FStartLink := plink;
  end;
  Inc(FCount);
end;

procedure TJvCCDList.DeleteFirst;
var
  plink : PCCDLink;
begin
  plink := FStartLink;
  if (plink <> nil) then
  begin
    FreeFromChain(plink);
    FreeLink(plink);
  end;
end;

procedure TJvCCDList.DeleteCCItemPtr(pitem : TJvCCItemPtr);
var
  plink : PCCDLink;
begin
  if (pitem=nil) then
    raise EJvCCRecordMgr.Create('Nil CCItemPtr');
  plink := PCCDLink(pitem);
  Dec(plink);
  FreeFromChain(plink);
  FreeLink(plink);
end;

procedure TJvCCDList.DeleteLast;
var
  plink : PCCDLink;
begin
  plink := FEndLink;
  if plink<>nil then
  begin
    FreeFromChain(plink);
    FreeLink(plink);
  end;
end;

function TJvCCDList.FirstItem :TJvCCItemPtr;
var {buf is assumed to be initialised}
  plink : PCCDLink;
begin
  plink := FStartLink;
  if (plink <> nil) then
    Inc(plink,1);
  Result := plink;
end;

function TJvCCDList.LastItem :TJvCCItemPtr;
var {buf is assumed to be initialised}
  plink : PCCDLink;
begin
  plink := FEndLink;
  if (plink <> nil) then
    Inc(plink,1);
  Result := plink;
end;

function TJvCCDList.NextItem(pitem : TJvCCItemPtr) : TJvCCItemPtr;
var {buf is assumed to be initialised}
  plink : PCCDLink;
begin
  plink := pitem;
  if (plink <> nil) then
  begin
    Dec(plink,1);
    plink := plink^.FNext;
    if (plink <> nil) then
      Inc(plink,1);
  end;
  Result := plink;
end;

function TJvCCDList.PriorItem(pitem : TJvCCItemPtr) : TJvCCItemPtr;
var {buf is assumed to be initialised}
  plink : PCCDLink;
begin
  plink := pitem;
  if (plink <> nil) then
  begin
    Dec(plink,1);
    plink := plink^.FPrior;
    if (plink <> nil) then
      Inc(plink,1);
  end;
  Result := plink;
end;


procedure TJvCCDList.FreeItem(var pitem : TJvCCItemPtr);
var
  plink : PCCDLink;
begin
  plink := pitem;
  if (plink = nil) then
    exit;
  Dec(plink,1);
  FreeFromChain(plink);
  FreeLink(plink);
end;

function TJvCCDList.AppendItem(const buf) : TJvCCItemPtr;
var
  plink : PCCDLink;
begin
  plink := NewLink(buf);
  if (FEndLink=nil) then
  begin
    FEndLink := plink;
    FStartLink := plink;
  end else
  begin
    plink.FPrior := FEndLink;
    FEndLink^.FNext := plink;
    FEndLink := plink;
  end;
  Inc(FCount);
  Inc(plink);
  Result := plink;
end;

function TJvCCDList.DeleteItem(const buf) : boolean;
var
  slink : PCCDLink;{test link}
  nlink : PCCDLink;{next}
  dlink : PCCDLink;{data}
begin
  slink := FStartLink;
  while (slink <> nil) do
  begin
    nlink := slink^.FNext;
    dlink := slink;
    Inc(dlink,1);
    if FRecMgr.CompareItems(dlink^,buf)=0 then
    begin
      FreeFromChain(slink);
      FreeLink(slink);
      Result := TRUE;
      exit;
    end;
    slink := nlink;
  end;
  Result := FALSE;
end;

procedure TJvCCDListIterator.NotifyDeleted(invalidlink,replacelink:PCCDLink);
begin
  if (invalidLink = FCurrLink) then
    FCurrLink := replacelink;
end;

constructor TJvCCDListIterator.Create(dlist:TJvCCDList);
begin
  inherited Create;
  FDList := dlist;
end;

function TJvCCDListIterator.First : boolean;
begin
  FCurrLink := FDList.FStartLink;
  Result  := (FCurrLink <> nil);
end;

function TJvCCDListIterator.Last : boolean;
begin
  FCurrLink := FDList.FEndLink;
  Result  := (FCurrLink <> nil);
end;

function TJvCCDListIterator.Next : boolean;
begin
  FCurrLink := FCurrLink.FNext;
  Result  := (FCurrLink <> nil);
end;

function TJvCCDListIterator.Prior : boolean;
begin
  FCurrLink := FCurrLink.FPrior;
  Result  := (FCurrLink <> nil);
end;

function TJvCCDListIterator.Data : Pointer;
var
  rp:PCCDLink;
begin
  rp := FCurrLink;
  if (rp <> nil) then
    Inc(rp,1);
  Result := rp;
end;


end.
