{-----------------------------------------------------------------------------
The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in compliance
with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/MPL-1.1.html

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either expressed or implied. See the License for
the specific language governing rights and limitations under the License.

The Original Code is: JvCCTree.PAS, released on 2000-12-01.

The Initial Developer of the Original Code is Michael Rynn.
Portions created by Michael Rynn are Copyright (C) 2000  Michael Rynn.
All Rights Reserved.

Contributor(s): ______________________________________.

Last Modified: 2000-12-01

You may retrieve the latest version of this file at the Project JEDI home page,
located at http://www.delphi-jedi.org

-----------------------------------------------------------------------------}
{$A+,B-,C+,D+,E-,F-,G+,H+,I+,J+,K-,L+,M-,N+,O+,P+,Q-,R-,S-,T-,U-,V+,W-,X+,Y+,Z1}

{$I JCL.INC}
unit JvCCTree;

interface
Uses JvCCRecrd, SysUtils;
type
  EJvCC = class(EJvCCRecordMgr);
type
  PCCTreeNode = ^TJvCCTreeNode;
  TJvCCTreeNode = record
    FLeft : PCCTreeNode;
    FRight : PCCTreeNode;
    FParent : PCCTreeNode;
    FColor : ShortInt;
  end;

  {nulls are used instead of sentinel node,
   code is adapted from the HP-STL for C++ stl_tree.h,
   now found on the SGI website
  }
  TJvCCTree = class(TJvCCBase)
  protected
    FNode : TJvCCTreeNode;
    FUnique : boolean;

    class function LeftMostNode(node : PCCTreeNode) : PCCTreeNode;
    class function RightMostNode(node : PCCTreeNode) : PCCTreeNode;
    class procedure Increment(var node : PCCTreeNode);
    class procedure Decrement(var node : PCCTreeNode);
    class procedure RotateLeft(x:PCCTreeNode; var root : PCCTreeNode);
    class procedure RotateRight(x:PCCTreeNode; var root : PCCTreeNode);
    class procedure InsertBalance(x:PCCTreeNode; var root : PCCTreeNode);
    class function DeleteBalance(z:PCCTreeNode;
      var root,leftmost, rightmost : PCCTreeNode) : PCCTreeNode;

    function NewNode : PCCTreeNode;
    procedure FreeNode(var plink : PCCTreeNode);
    procedure AssignNodeData(node : PCCTreeNode; const buf);
    procedure FetchNodeData(node : PCCTreeNode; var buf);
    function SearchForNode(
      data:Pointer; var node: PCCTreeNode; iskey:boolean) : boolean;
    function LocateNodeData(
      data : Pointer;
      var startItem : TJvCCItemPtr;
      direction : TJvCCItemSearch;
      iskey:boolean) : boolean;

    procedure PrecNode(var node : PCCTreeNode);
    procedure SuccNode(var node : PCCTreeNode);
    property RootNode : PCCTreeNode read FNode.FParent;
    property FrontNode : PCCTreeNode read FNode.FLeft;
    property BackNode : PCCTreeNode read FNode.FRight;
  public
    constructor CreateMgr(reclass : TJvCCRecordMgrClass); override;
    function LocateKeyPtr(
      key : Pointer;
      var startItem : TJvCCItemPtr;
      direction : TJvCCItemSearch) : boolean; override;
    function LocateItemPtr(
      const rec;
      var startItem : TJvCCItemPtr;
      direction : TJvCCItemSearch) : boolean; override;
    function AppendItem(const buf) : TJvCCItemPtr; override;
    function DeleteItem(const buf) : boolean; override;
    function InsertNode(const buf; var item: TJvCCItemPtr) : boolean;
    procedure FreeItem(var pitem : TJvCCItemPtr);  override;
    function FirstItem : TJvCCItemPtr; override;
    function LastItem : TJvCCItemPtr; override;
    function NextItem(pitem : TJvCCItemPtr) : TJvCCItemPtr; override;
    function PriorItem(pitem : TJvCCItemPtr) : TJvCCItemPtr; override;
    procedure Clear; override;
    property Unique : boolean read FUnique write FUnique;
  end;

  TJvCCTreeIterator = class(TObject)
  protected
    FTree : TJvCCTree;
    FNode : PCCTreeNode;
  public
    constructor Create(tree:TJvCCTree);
    function First : boolean;
    function Last : boolean;
    function Next : boolean;
    function Prior : boolean;
    function DataPtr : Pointer;
    procedure GetDataCopy(var buf);
  end;

implementation

const
  NBLACK = 0;
  NRED = 1;

class function TJvCCTree.LeftMostNode(node : PCCTreeNode) : PCCTreeNode;
begin
  while (node^.FLeft <> nil) do
    node := node^.FLeft;
  Result := node;
end;

class function TJvCCTree.RightMostNode(node : PCCTreeNode) : PCCTreeNode;
begin
  while (node^.FRight <> nil) do
    node := node^.FRight;
  Result := node;
end;

class procedure TJvCCTree.Increment(var node : PCCTreeNode);
var
  p : PCCTreeNode;
begin
  if (node^.FRight <> nil) then
  begin
    node := node^.FRight;
    while (node^.FLeft <> nil) do
      node := node^.FLeft;
  end else
  begin
    p := node^.FParent;
    while ((node<>p^.FParent) and (node <> p^.FLeft)) do
    begin
      node := p;
      p := node^.FParent;
    end;
    if (node^.FRight <> p) then
      node := p;
  end;
end;

class procedure TJvCCTree.Decrement(var node : PCCTreeNode);
var
  p : PCCTreeNode;
begin
  if (node^.FColor = NRED) and (
    node^.FParent^.FParent = node) then
    node := node^.FRight
  else
  begin
    p := node^.FLeft;
    if (p <> nil) then
    begin
      while( p^.FRight <> nil) do
        p := p^.FRight;
      node := p;
    end else
    begin
      p := node^.FParent;
      while( node = p^.FLeft) do
      begin
        node := p;
        p := p^.FParent;
      end;
      node := p;
    end;
  end;
end;


procedure TJvCCTree.PrecNode(var node : PCCTreeNode);
begin
  if (node = nil) then
  begin
    node := PCCTreeNode(LastItem);
    Dec(node,1);
  end else
  begin
    Decrement(node);
    if (node = @FNode) then
      node := nil;
  end;
end;

procedure TJvCCTree.SuccNode(var node : PCCTreeNode);
begin
  if (node = nil) then
  begin
    node := PCCTreeNode(FirstItem);
    Dec(node,1);
  end else
  begin
    Increment(node);
    if (node = @FNode) then
      node := nil;
  end;
end;


class procedure TJvCCTree.RotateLeft(x:PCCTreeNode; var root : PCCTreeNode);
var
  p : PCCTreeNode;
  y : PCCTreeNode;
begin
  y := x^.FRight;
  p := y^.FLeft;
  x^.FRight :=  p;
  if (p<>nil) then
    p^.FParent := x;
  p := x^.FParent;
  y^.FParent := p;
  if (root = x) then
    root := y
  else
  if (x = p^.FLeft) then
    p^.FLeft := y
  else
    p^.FRight := y;
  y^.FLeft := x;
  x^.FParent := y;
end;

class procedure TJvCCTree.RotateRight(x:PCCTreeNode; var root : PCCTreeNode);
var
  y : PCCTreeNode;
  p : PCCTreeNode;
begin
  y := x^.FLeft;
  p := y^.FRight;
  x^.FLeft := p;
  if (p <> nil) then
    p^.FParent := x;
  p := x^.FParent;
  y^.FParent := p;
  if (root = x) then
    root := y
  else if (x = p^.FLeft) then
    p^.FLeft := y
  else
    p^.FRight := y;
  y^.FRight := x;
  x^.FParent := y;
end;


class procedure TJvCCTree.InsertBalance(x:PCCTreeNode ; var root : PCCTreeNode);
var
  p : PCCTreeNode;
  y : PCCTreeNode;
begin
  x^.FColor := NRED;
  while (x <> root) do
  begin
    p := x^.FParent;
    if (p^.FColor = NBLACK) then
      break;
    if (p = p^.FParent^.FLeft) then
    begin
      y := p^.FParent^.FRight;
      if (y <> nil) and (y^.FColor=NRED) then
      begin
        p^.FColor := NBLACK;
        y^.FColor := NBLACK;
        x := p^.FParent;
        x^.FColor := NRED;
      end else
      begin
        if (x = p^.FRight) then
        begin
          x := p;
          RotateLeft(x, root);
          p := x^.FParent;
        end;
        p^.FColor := NBLACK;
        p := p^.FParent;
        p^.FColor := NRED;
        RotateRight(p,root);
      end;
    end else
    begin
      y := p^.FParent^.FLeft;
      if (y <> nil) and (y^.FColor=NRED) then
      begin
        p^.FColor := NBLACK;
        y^.FColor := NBLACK;
        x := p^.FParent;
        x^.FColor := NRED;
      end else
      begin
        if (x = p^.FLeft) then
        begin
          x := p;
          RotateRight(x, root);
          p := x^.FParent;
        end;
        p^.FColor := NBLACK;
        p := p^.FParent;
        p^.FColor := NRED;
        RotateLeft(p, root);
      end;
    end;
  end;
  root^.FColor := NBLACK;
end;


class function TJvCCTree.DeleteBalance(z:PCCTreeNode; var root,leftmost, rightmost : PCCTreeNode)
 : PCCTreeNode;
var
  y : PCCTreeNode;
  x : PCCTreeNode;
  p : PCCTreeNode;
  xp : PCCTreeNode;
  w : PCCTreeNode;
  lostcolor : integer;
begin
  y := z;
  if (y^.FLeft = nil) then
    x := y^.FRight
  else
  begin
    if (y^.FRight = nil) then
      x := y^.FLeft
    else
    begin
      y := y^.FRight;
      while(y^.FLeft <> nil) do
        y := y^.FLeft;
      x := y^.FRight;
    end;
  end;
  if (y <> z) then
  begin
    {relink y in place of z}
    z^.FLeft^.FParent := y;
    y^.FLeft := z^.FLeft;
    if (y <> z^.FRight) then
    begin
      xp := y^.FParent;
      if (x<>nil) then
        x^.FParent := xp;
      xp^.FLeft := x;
      y^.FRight := z^.FRight;
      z^.FRight^.FParent := y;
    end
    else
      xp := y;
    if (root = z) then
      root := y
    else
    begin
      p := z^.FParent;
      if (p^.FLeft = z) then
        p^.FLeft := y
      else
        p.FRight := y;
    end;
    y^.FParent := z^.FParent;
    lostcolor := y^.FColor;
    y^.FColor := z^.FColor;
    z^.FColor := lostcolor;
    y := z;
  end else
  begin {y=z to start with}
    xp := y^.FParent;
    if (x<>nil) then
      x^.FParent := xp;
    if (root = z) then
      root := x
    else
    begin
      p := z^.FParent;
      if (p^.FLeft=z) then
        p^.FLeft := x
      else
       p^.FRight := x;
    end;
    if (leftmost=z) then
      if (z^.FRight=nil) then
        leftmost := z^.FParent
      else
        leftmost := LeftMostNode(x);
    if (rightmost=z) then
      if (z^.FLeft=nil) then
        rightmost := z^.FParent
      else
        rightmost := RightMostNode(x);
  end;
  if (y^.FColor<>NRED) then
  begin
    while (x <> root) and ((x=nil) or (x^.FColor=NBLACK)) do
    begin
      if (x=xp^.FLeft) then
      begin
        w  := xp^.FRight;
        if (w^.FColor=NRED) then
        begin
          w^.FColor := NBLACK;
          xp^.FColor := NRED;
          RotateLeft(xp,root);
          w := xp^.FRight;
        end;
        if ((w^.FLeft=nil) or (w^.FLeft^.FColor=NBLACK)) and
           ((w^.FRight=nil) or (w^.FRight^.FColor=NBLACK)) then
        begin
          w^.FColor := NRED;
          x := xp;
          xp := x^.FParent;
        end else
        begin
          if  ((w^.FRight=nil) or (w^.FRight^.FColor=NBLACK)) then
          begin
            w^.FLeft^.FColor := NBLACK;
            w^.FColor := NRED;
            RotateRight(w,root);
            w := xp^.FRight;
          end;
          w^.FColor := xp^.FColor;
          xp^.FColor := NBLACK;
          w^.FRight^.FColor := NBLACK;
          RotateLeft(xp,root);
          break;
        end;
      end else
      begin
        w := xp^.FLeft;
        if (w^.FColor=NRED) then
        begin
          w^.FColor := NBLACK;
          xp^.FColor := NRED;
          RotateRight(xp,root);
          w := xp^.FLeft;
        end;
        if ((w^.FRight=nil) or (w^.FRight^.FColor=NBLACK)) and
           ((w^.FLeft=nil) or (w^.FLeft^.FColor=NBLACK)) then
        begin
          w^.FColor := NRED;
          x := xp;
          xp := x^.FParent;
        end else
        begin
          if  ((w^.FLeft=nil) or (w^.FLeft^.FColor=NBLACK)) then
          begin
            w^.FRight^.FColor := NBLACK;
            w^.FColor := NRED;
            RotateLeft(w,root);
            w := xp^.FLeft;
          end;
          w^.FColor := xp^.FColor;
          xp^.FColor := NBLACK;
          w^.FLeft^.FColor := NBLACK;
          RotateRight(xp,root);
          break;
        end;
      end;
    end;
    if (x <> nil) then
      x^.FColor := NBLACK;
  end;
  Result := y;
end;

constructor TJvCCTree.CreateMgr(reclass : TJvCCRecordMgrClass);
begin
  inherited CreateMgr(reclass);
end;

function TJvCCTree.NewNode : PCCTreeNode;
var
  linksize : integer;
  p : PCCTreeNode;
begin
  linksize := sizeof(TJvCCTreeNode) + FRecMgr.ItemSize;
  GetMem(p,linksize);
  p^.FParent := nil;
  p^.FLeft := nil;
  p^.FRight := nil;
  p^.FColor := NBLACK;
  Result := p;
  Inc(p,1);
  if FRecMgr.InitItems then
    FRecMgr.ConstructItem(p^);
end;

procedure TJvCCTree.FetchNodeData(node : PCCTreeNode; var buf);
var
  dlink : PCCTreeNode;{data}
begin
  dlink := node;
  Inc(dlink,1);
  FRecMgr.AssignItem(buf,dlink^);
end;

procedure TJvCCTree.FreeItem(var pitem : TJvCCItemPtr);
var
  x : PCCTreeNode;
begin
  if (pitem = nil) then
    exit;
  x := pitem;
  pitem := nil;
  Dec(x,1);
  x := DeleteBalance(x,FNode.FParent,FNode.FLeft,FNode.FRight);
  Dec(FCount);
  FreeNode(x);
end;

procedure TJvCCTree.AssignNodeData(node : PCCTreeNode; const buf);
var
  dlink : PCCTreeNode;{data}
begin
  dlink := node;
  Inc(dlink,1);
  FRecMgr.AssignItem(dlink^,buf);
end;

procedure TJvCCTree.FreeNode(var plink : PCCTreeNode);
var
  linksize : integer;
  dlink : PCCTreeNode;{data}
begin
  if (plink = nil) then
    exit;
  dlink := plink;
  Inc(dlink,1);
  if FRecMgr.InitItems then
    FRecMgr.DestructItem(dlink^);
  linksize := sizeof(TJvCCTreeNode) + FRecMgr.ItemSize;
  FreeMem(plink,linksize);
  plink := nil;
end;

function TJvCCTree.AppendItem(const buf) : TJvCCItemPtr;
var
  item : TJvCCItemPtr;
begin
  InsertNode(buf,item);
  Result := item;
end;

function TJvCCTree.InsertNode(const buf; var item : TJvCCItemPtr) : boolean;
var
  x : PCCTreeNode;
  p : PCCTreeNode;
  z : PCCTreeNode;
  nd : PCCTreeNode;
  dres : integer;
begin
  x := RootNode;
  p := @FNode;
  dres := -1;
  while (x <> nil) do
  begin
    p := x;
    nd := x;
    Inc(nd,1);
    dres := FRecMgr.CompareItems(buf,nd^);
    if (dres < 0) then
    begin
      x := x^.FLeft;
    end else
    if (dres > 0) then
    begin
      x := x^.FRight
    end else
    begin
      if FUnique then
      begin
        item := TJvCCItemPtr(nd);
        Result := FALSE;
        exit;
      end;
      break;
    end;
  end;
  z := NewNode;
  AssignNodeData(z,buf);
  if (p = @FNode) or (x <> nil) or (dres<0) then
  begin
    p^.FLeft := z;
    if (p=@FNode) then
    begin
      FNode.FParent := z;
      FNode.FRight := z;
    end else
    if (p = FNode.FLeft) then
      FNode.FLeft := z;
  end else
  begin
    p^.FRight := z;
    if (p = FNode.FRight) then
      FNode.FRight := z
  end;
  z^.FParent := p;
  InsertBalance(z,FNode.FParent);
  Inc(FCount);
  Inc(z,1);
  item := z;
  Result := TRUE;
end;

function TJvCCTree.LocateNodeData(
      data:Pointer; var startItem : TJvCCItemPtr;
      direction : TJvCCItemSearch; iskey:boolean) : boolean;
var
  match : integer;
  node : PCCTreeNode;
begin
  node := startItem;
  if (node <> nil) then
  begin
    if iskey then
      match := FRecMgr.CompareKey(node^,data)
    else
      match := FRecMgr.CompareItems(node^,data^);
    if (match = 0) then
    begin
      Result := TRUE;
      exit;
    end;
    startItem := nil;
    CASE direction of
      its_forward:
        if (match > 0) then  {already too high}
        begin
          Result := FALSE;
          exit;
        end;
      its_backward:
        if (match < 0) then {already too low}
        begin
          Result := FALSE;
          exit;
        end;
    END;{CASE}
    // search from Root again for key
  end;
  Result := SearchForNode(data,node,iskey);
  if (node <> nil) then
  begin
    Inc(node,1);
    startItem := node;
  end;
end;

function TJvCCTree.LocateItemPtr(
  const rec; var startItem : TJvCCItemPtr;
  direction : TJvCCItemSearch) : boolean;
begin
  Result := LocateNodeData(@rec,startItem,direction,FALSE);
end;

function TJvCCTree.LocateKeyPtr(
  key : Pointer; var startItem : TJvCCItemPtr;
  direction : TJvCCItemSearch) : boolean;
begin
  Result := LocateNodeData(key,startItem,direction,TRUE);
end;


function TJvCCTree.SearchForNode(
  data:Pointer;
  var node:PCCTreeNode;
  iskey : boolean) : boolean;
var
  match : integer;
  nd : PCCTreeNode;
begin
  nd := RootNode;
  while (nd <> nil) do
  begin
    node := nd;
    Inc(nd,1);
    if iskey then
      match := FRecMgr.CompareKey(nd^,data)
    else
      match := FRecMgr.CompareItems(nd^,data^);
    Dec(nd,1);
    if (match > 0) then
      nd := nd^.FLeft
    else if (match < 0) then
      nd := nd^.FRight
    else
    begin
      Result := TRUE;
      exit;
    end;
  end;
  Result := FALSE;
end;


function TJvCCTree.DeleteItem(const buf) : boolean;
var
  node : PCCTreeNode;
begin
  Result := SearchForNode(@buf,node,FALSE);
  if Result then
  begin
    Inc(node,1);
    FreeItem(TJvCCItemPtr(node));
  end;
end;


function TJvCCTree.FirstItem : TJvCCItemPtr;
var
  node : PCCTreeNode;
begin
  if (RootNode <> nil) then
    node := LeftMostNode(RootNode)
  else
    node := nil;
  if (node = nil) then
  begin
    Result := nil;
  end else
  begin
    Inc(node,1);
    Result := node;
  end;
end;

function TJvCCTree.LastItem : TJvCCItemPtr;
var
  node : PCCTreeNode;
begin
  if(RootNode <> nil) then
    node := RightMostNode(RootNode)
  else
    node := nil;
  if (node = nil) then
  begin
    Result := nil;
  end else
  begin
    Inc(node,1);
    Result := node;
  end;
end;

function TJvCCTree.NextItem(pitem : TJvCCItemPtr) : TJvCCItemPtr;
var
  node : PCCTreeNode;
begin
  node := pitem;
  if (node = nil) then
  begin
    Result := nil;
    exit;
  end;
  Dec(node,1);
  SuccNode(node);
  if (node <> nil) then
  begin
    Inc(node,1);
    Result := node;
  end else
  begin
    Result := nil;
  end;
end;


function TJvCCTree.PriorItem(pitem : TJvCCItemPtr) : TJvCCItemPtr;
var
  node : PCCTreeNode;
begin
  node := pitem;
  if (node = nil) then
  begin
    Result := nil;
    exit;
  end;
  Dec(node,1);
  PrecNode(node);
  if (node <> nil) then
  begin
    Inc(node,1);
    Result := node;
  end else
  begin
    Result := nil;
  end;
end;


constructor TJvCCTreeIterator.Create(tree:TJvCCTree);
begin
  inherited Create;
  FTree := tree;
  FNode := nil;
end;

function TJvCCTreeIterator.First : boolean;
begin
  if (FTree.RootNode <> nil) then
  begin
    FNode := FTree.LeftMostNode(FTree.RootNode);
    Result := (FNode <> nil);
  end
  else
    Result := FALSE;
end;

function TJvCCTreeIterator.Last : boolean;
begin
  if (FTree.RootNode <> nil) then
  begin
    FNode := FTree.RightMostNode(FTree.RootNode);
    Result := (FNode <> nil);
  end else
    Result := FALSE;
end;

function TJvCCTreeIterator.Next : boolean;
begin
  if FNode <> nil then
  begin
    FTree.SuccNode(FNode);
    Result := (FNode <> nil);
  end
  else
    Result := FALSE;
end;

function TJvCCTreeIterator.Prior : boolean;
begin
  if (FNode <> nil) then
  begin
    FTree.PrecNode(FNode);
    Result := (FNode <> nil);
  end
  else
    Result := FALSE;
end;

function TJvCCTreeIterator.DataPtr : Pointer;
var
  dnode : PCCTreeNode;
begin
  if (FNode = nil) then
    Result := nil
  else
  begin
    dnode := FNode;
    Inc(dnode,1);
    Result := dnode;
  end;
end;

procedure TJvCCTree.Clear;
var
  ip : TJvCCItemPtr;
begin
  ip := FirstItem;
  while (ip <> nil) do
  begin
    FreeItem(ip);
    ip := FirstItem;
  end;
end;

procedure TJvCCTreeIterator.GetDataCopy(var buf);
begin
  if (FNode = nil) then
    raise EJvCC.Create('Nil node pointer')
  else
    FTree.FetchNodeData(FNode,buf);
end;
end.
