{-----------------------------------------------------------------------------
The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in compliance
with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/MPL-1.1.html

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either expressed or implied. See the License for
the specific language governing rights and limitations under the License.

The Original Code is: JvRnnDB_BDE.PAS, released on 2000-12-01.

The Initial Developer of the Original Code is Michael Rynn.
Portions created by Michael Rynn are Copyright (C) 2000  Michael Rynn.
All Rights Reserved.

Contributor(s): ______________________________________.

Last Modified: 2000-12-01

You may retrieve the latest version of this file at the Project JEDI home page,
located at http://www.delphi-jedi.org

-----------------------------------------------------------------------------}
{$A+,B-,C+,D+,E-,F-,G+,H+,I+,J+,K-,L+,M-,N+,O+,P+,Q-,R-,S-,T-,U-,V+,W-,X+,Y+,Z1}

{$I JCL.INC}
unit JvRnnDB_BDE;

interface
uses Classes,JvRnnGenDB,DB,DBTables,JvRnnDBScript;

type
  TJvRnnDBLinkBDE = class(TJvRnnGenDBLink)
  protected
    FDatabase : TDatabase;
    FQuery : TQuery;
  public
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
    procedure SetConnected(islive : boolean); override;
    function GetConnected : boolean; override;
    function  ReturnDataset(slist:TStrings) : TJvRnnDatasetLink; override;
    procedure ExecuteStd(cmd:TStrings);  override;
  end;

  TJvRnnDatasetBDE = class(TJvRnnDatasetLink)
  protected
    procedure SetSQL(slist:TStrings); override;
    function GetSQL : TStrings; override;
  public
    constructor Create(dblink:TJvRnnGenDBLink); override;
    destructor Destroy; override;
  end;

procedure Register;

implementation
uses JvRnnStrUt,Sysutils;

const
  rdb_NILDataset = 0;
  kAliasName = 'AliasName';
  kDriverName = 'DriverName';
  kEmptyStr = '';

procedure RDBError(i:Integer);
var
  s : string;
begin
  CASE i of
  rdb_NILDataset :
    s := 'NIL dataset';
  END;
  raise EJvRnnDB.Create(s);
end;

constructor TJvRnnDBLinkBDE.Create(AOwner:TComponent);
begin
  inherited;
  FDatabase := TDatabase.Create(SELF);
  FQuery := TQuery.Create(SELF);
end;

destructor TJvRnnDBLinkBDE.Destroy;
begin
  FQuery.Free;
  FDatabase.Free;
  inherited;
end;

procedure TJvRnnDBLinkBDE.ExecuteStd(cmd : TStrings);
begin
  FQuery.sql.Assign(cmd);
  if cmd.count = 0 then
    exit;
  FQuery.Prepare;
  FQuery.ExecSQL;
  FQuery.close;
end;

procedure TJvRnnDBLinkBDE.SetConnected(islive : boolean);
var
  i : integer;
  slist : TStrings;
begin
  if (FDatabase.Connected = islive) then
    exit;
  if islive then
  begin
    slist := TStringList.Create;
    try
      slist.Assign(FConnectLink.Params);
      i := slist.IndexOfName(kAliasName);
      if (i >= 0) then
      begin
        FDatabase.AliasName := GetStrValue(slist[i]);
        slist.Delete(i);
      end;
      i := slist.IndexOfName(kDriverName);
      if (i >= 0) then
      begin
        FDatabase.DriverName := GetStrValue(slist[i]);
        slist.Delete(i);
      end;
      i := slist.IndexOfName(kDatabase);
      if (i >= 0) then
      begin
        FDatabase.DatabaseName := GetStrValue(slist[i]);
        slist.Delete(i);
      end;
      FDatabase.Params := slist;
    finally
      slist.free;
    end;
  end;
  FDatabase.Connected := islive;
end;

function TJvRnnDBLinkBDE.GetConnected : boolean;
begin
  Result := FDatabase.Connected;
end;

function  TJvRnnDBLinkBDE.ReturnDataset(slist:TStrings) : TJvRnnDatasetLink;
var
  link : TJvRnnDatasetBDE;
begin
  Result := TJvRnnDatasetBDE.Create(SELF);
  try
    if (slist <> nil) then
    begin
      link.SetSQL(slist);
      link.Dataset.Open;
    end;
    FDatasetLinks.AppendItem(link);
  except
    FreeAndNil(Result);
    raise;
  end;
end;

constructor TJvRnnDatasetBDE.Create(dblink:TJvRnnGenDBLink);
begin
  inherited Create(dblink);
  FDataset := TQuery.Create(dblink);
end;

destructor TJvRnnDatasetBDE.Destroy;
begin
  FDataset.Free;
  FDataset := nil;
  inherited;
end;

function TJvRnnDatasetBDE.GetSQL : TStrings;
begin
  if FDataset = nil then
    RDBError(rdb_nilDataset);
  Result := TQuery(FDataset).SQL;
end;

procedure TJvRnnDatasetBDE.SetSQL(slist:TStrings);
begin
  if FDataset = nil then
    RDBError(rdb_nilDataset);
  TQuery(FDataset).SQL := slist;
end;

procedure SetupDB_BDE;
begin
  TJvGenDBScript.RegisterDBLink('BDE',TJvRnnDBLinkBDE);
end;

procedure Register;
begin
  RegisterComponents('RnnLib',[TJvRnnDBLinkBDE]);
end;

initialization
  SetupDB_BDE;
end.
