{-----------------------------------------------------------------------------
The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in compliance
with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/MPL-1.1.html

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either expressed or implied. See the License for
the specific language governing rights and limitations under the License.

The Original Code is: JvRnnGenDB.PAS, released on 2000-12-01.

The Initial Developer of the Original Code is Michael Rynn.
Portions created by Michael Rynn are Copyright (C) 2000  Michael Rynn.
All Rights Reserved.

Contributor(s): ______________________________________.

Last Modified: 2000-12-01

You may retrieve the latest version of this file at the Project JEDI home page,
located at http://www.delphi-jedi.org

-----------------------------------------------------------------------------}
{$A+,B-,C+,D+,E-,F-,G+,H+,I+,J+,K-,L+,M-,N+,O+,P+,Q-,R-,S-,T-,U-,V+,W-,X+,Y+,Z1}

{$I JCL.INC}
unit JvRnnGenDB;

interface
uses
  JvCCRecrd,JvCCArray,JvCCLists,DB,DBTables,Classes,SysUtils,JvRnnScCst;
const
  kUserName = 'UserName';
  kPassword = 'Password';
  kDatabase = 'Database';

type
  TDBLinkOp = (lop_append, lop_update, lop_delete, lop_insertnew);
  TDBLinkOpSet = set of TDBLinkOp;

  TDBConnectLink = class(TObject)
  protected
    FConnectName : string;
    FParams : TStrings;
  public
    constructor Create; virtual;
    destructor Destroy; override;
    procedure SetParams(slist:TStrings);
    property Params : TStrings read FParams write SetParams;

  end;

  TJvRnnGenDBLink = class;
  TJvRnnDatasetLink = class(TObject)
  protected
    FDBLink : TJvRnnGenDBLink;
    FDataset : TDataset;
    FDSName : string;
    procedure SetSQL(slist:TStrings); virtual;abstract;
    function GetSQL : TStrings;  virtual;abstract;
    procedure ForgetDataset;
  public
    constructor Create(dblink:TJvRnnGenDBLink);  virtual;
    destructor Destroy; override;
    property SQL : TStrings read GetSQL write SetSQL;
    property Dataset : TDataset read FDataset;
    property DSName : string read FDSName write FDSName;
    property DBLink : TJvRnnGenDBLink read FDBLink;
  end;

  TDatasetList = class(TJvCCObjArray)
  public
    constructor Create; 
  end;

  TJvRnnGenDBLink = class(TComponent)
  protected
    FDatasetLinks : TDatasetList;
    FConnectLink : TDBConnectLink;
    FScriptor : TJvRnnCustomScript;
    FLinkType : string;

    procedure SetLinkName(const link_name:string);
    function GetLinkName : string;
    procedure FreeDatasetLink(ip:TJvCCItemPtr;data:Pointer);
    procedure ForgetDatasetLink(ip:TJvCCItemPtr;data:Pointer);
  public
    class function DBLinkDescription : string; virtual;
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
    function TablesExist(tablist:TStrings) : boolean; virtual; abstract;
    procedure SetConnected(islive : boolean); virtual; abstract;
    function GetConnected : boolean; virtual; abstract;
    procedure ExecSQLStr(const sql : string);
    procedure TableFromFieldList(
      const new_name : string;
      table_fields, key_fields : TJvCCObjArray); virtual;

    procedure AppendUpdateFromDataset(
      ds : TDataset;
      const dest_table_name : string;
      table_fields, key_fields: TJvCCObjArray;
      opSet : TDBLinkOpSet);  virtual;

    function ReturnDataset(slist:TStrings) :
      TJvRnnDatasetLink;  virtual; abstract;

    function GetSQLFieldType(
      flddef : TFieldDef) : string; virtual;
      
    procedure ExecuteStd(cmd:TStrings);  virtual; abstract;
    procedure ForgetDatasets;

    procedure SetScriptor(scriptor:TJvRnnCustomScript);
    
    function RecordsExist(
      const select_sql:string;  params : variant) : boolean; virtual; abstract;
    procedure UpdateRecords(
      const update_sql:string; params : variant); virtual; abstract;
    procedure InsertRecords(
      const insert_sql:string; params : variant); virtual; abstract;

    procedure BeginTransaction; virtual; abstract;
    procedure Commit; virtual; abstract;
    procedure Rollback; virtual; abstract;
    procedure CommitRetaining; virtual; abstract;

    property LinkType : string read FLinkType write FLinkType;
    property Scriptor : TJvRnnCustomScript read FScriptor;
    property LinkName : string read GetLinkName write SetLinkName;
    property ConnectLink : TDBConnectLink read FConnectLink;
    property Connected : boolean
      read GetConnected write SetConnected;
    property DSList : TDatasetList read FDatasetLinks;
  end;


  TJvRnnGenDBLinkClass = class of TJvRnnGenDBLink;

  EJvRnnDB = class(Exception);


implementation

class function TJvRnnGenDBLink.DBLinkDescription : string;
begin
  Result := 'Override this generic link description';
end;

function DSObjCompare(obj1,obj2:TObject):integer;
begin
  Result := CompareStr(TJvRnnDatasetLink(obj1).DSName,
    TJvRnnDatasetLink(obj2).DSName);
end;

function DSKeyCompare(obj:TObject;key:Pointer):integer;
begin
  Result := CompareStr(TJvRnnDatasetLink(obj).DSName,
    PString(key)^);
end;

constructor TJvRnnGenDBLink.Create;
begin
  FDatasetLinks := TDatasetList.Create;
  FConnectLink := TDBConnectLink.Create;
end;

procedure TJvRnnGenDBLink.FreeDatasetLink(
  ip:TJvCCItemPtr;
  data:Pointer);
var
  link:TJvRnnDatasetLink;
begin
  link := TJvRnnDatasetLink(ip^);
  link.FDBLink := nil;
  link.free;
end;

procedure TJvRnnGenDBLink.ForgetDatasetLink(
  ip:TJvCCItemPtr;
  data:Pointer);
var
  link:TJvRnnDatasetLink;
begin
  link := TJvRnnDatasetLink(ip^);
  link.ForgetDataset;
end;

procedure TJvRnnGenDBLink.SetLinkName(const link_name:string);
begin
  FConnectLink.FConnectName := link_name;
end;


procedure TJvRnnGenDBLink.SetScriptor(scriptor:TJvRnnCustomScript);
begin
  FScriptor := scriptor;
end;

function TJvRnnGenDBLink.GetLinkName : string;
begin
  Result := FConnectLink.FConnectName;
end;

procedure TJvRnnGenDBLink.ForgetDatasets;
begin
  DSList.CallActionMethod(SELF.ForgetDatasetLink, nil);
end;

procedure TJvRnnGenDBLink.TableFromFieldList(
  const new_name : string;
  table_fields, key_fields : TJvCCObjArray);
var
  sql : string;
  sqlspec : string;
  i : integer;
  fld : TFieldDef;
  //iskey : boolean;
begin
  sql := Format('create table %s (',[new_name]);
  for i := 0 to table_fields.count-1 do
  begin
    table_fields.GetItem(i,fld);
    //iskey := (key_fields <> nil) and (key_fields.IndexOfItem(fld) >= 0);
    sqlspec := GetSQLFieldType(fld);
    if (i > 0) then
      sql := sql + ',';
    sql := sql + sqlspec;
  end;
  {borland style primary key sql spec}
  if( key_fields <> nil) then
  begin
    sql := sql + ', primary key (';
    for i := 0 to key_fields.count-1 do
    begin
      key_fields.GetItem(i,fld);
      if (i > 0) then
        sql := sql + ',';
      sql := sql + fld.Name;
    end;
    sql := sql + ')';
  end;
  sql := sql + ')';
  ExecSQLStr(sql);
end;

function GenericInsertSQL(
  ds : TDataset;
  const dest_table_name : string;
  table_fields, key_fields: TJvCCObjArray) : string;
var
  i : integer;
  flddef : TFieldDef;
begin
  Result := Format('insert into %s (',[dest_table_name]);
  for i := 0 to table_fields.count - 1 do
  begin
    table_fields.GetItem(i,flddef);
    if (i > 0) then
      Result := Result + ',';
    Result := Result + flddef.Name;
  end;
  Result := Result + ') values (';
  for i := 0 to table_fields.count - 1 do
  begin
    table_fields.GetItem(i,flddef);
    if (i > 0) then
      Result := Result + ',';
    Result := Format('%s :%s',[Result,flddef.Name]);
  end;
  Result := Result + ')';
end;

function GenericUpdateSQL(
  ds : TDataset;
  const dest_table_name : string;
  table_fields, key_fields, update_fields: TJvCCObjArray) : string;
var
  i : integer;
  k : integer;
  flddef : TFieldDef;
begin
  if (key_fields = nil) then
     raise EJvRnnScript.create('Updates must have key fields');
  Result := Format('update %s set ',[dest_table_name]);
  {subtract any key_fields from table_fields}
  k := 0;
  for i := 0 to table_fields.count - 1 do
  begin
    table_fields.GetItem(i,flddef);
    if (key_fields.IndexOfItem(flddef) < 0) then
    begin
      Inc(k);
      if (k > 1) then
        Result := Result + ',';
      update_fields.AppendItem(flddef);
      Result := Format('%s %s = :%s',[Result, flddef.Name,flddef.Name]);
    end;
  end;
  Result := Result + ' where ';
  for i := 0 to key_fields.count - 1 do
  begin
    key_fields.GetItem(i,flddef);
    if (i > 0) then
      Result := Result + ' and ';
    Result := Format('%s %s = :%s',[Result,flddef.Name,flddef.Name]);
    update_fields.AppendItem(flddef);
  end;
end;

function GenericSelectSQL(
  ds : TDataset;
  const sel_table_name : string;
  table_fields, key_fields: TJvCCObjArray) : string;
var
  i : integer;
  flddef : TFieldDef;
begin
  {subtract any key_fields from table_fields}
  Result := '';
  for i := 0 to table_fields.count - 1 do
  begin
    table_fields.GetItem(i,flddef);
    if (i > 0) then
        Result := Result + ',';
    Result := Result + flddef.Name;
  end;
  Result := Format('select %s from %s where ',[Result,sel_table_name]);
  for i := 0 to key_fields.count - 1 do
  begin
    key_fields.GetItem(i,flddef);
    if (i > 0) then
      Result := Result + ' and ';
    Result := Format('%s %s = :%s',[Result,flddef.Name,flddef.Name]);
  end;
end;

procedure TJvRnnGenDBLink.AppendUpdateFromDataset(
  ds : TDataset;
  const dest_table_name : string;
  table_fields, key_fields: TJvCCObjArray;
  opSet : TDBLinkOpSet);
var
  insert_sql : string;
  update_sql : string;
  select_sql : string;
  i : integer;
  flddef : TFieldDef;
  params : variant;
  highix : integer;
  dsfield : TField;
  update_fields : TJvCCObjArray;
  exists : boolean;
begin
  ds.First;
  if (ds.EOF) then
    exit;
  {insert uses table_fields}
  {select uses key_fields}
  {update uses update_fields}
  update_fields := TJvCCObjArray.CreateMgr(TJvCCObjRecMgr);
  try
    insert_sql := GenericInsertSQL(
      ds, dest_table_name, table_fields,
      key_fields);
    update_sql := GenericUpdateSQL(
      ds, dest_table_name, table_fields,
      key_fields, update_fields);
    select_sql := GenericSelectSQL(
      ds, dest_table_name, table_fields, key_fields);
    highix := table_fields.count + key_fields.count-1;
    params := VarArrayCreate( [0, highix],varVariant);
    while (not ds.eof) do
    begin
      {if appending only, assume record does not exist}
      if not (opSet = [lop_append]) then
      begin
        for i := 0 to key_fields.count-1 do
        begin
          key_fields.GetItem(i,flddef);
          dsfield := ds.Fields[flddef.Index];
          params[i] := dsfield.AsVariant;
        end;
        exists :=  RecordsExist(select_sql, params);
      end else
        exists := FALSE;

      if ((not exists) and (lop_insertnew in opSet)) OR
        (opSet = [lop_append]) then
      begin
        // insert
        for i := 0 to table_fields.count-1 do
        begin
          table_fields.GetItem(i,flddef);
          dsfield := ds.Fields[flddef.Index];
          params[i] := dsfield.AsVariant;
        end;
        InsertRecords(insert_sql, params);
      end;
      if (exists) AND (lop_update in opSet) then
      begin
        // update
        for i := 0 to update_fields.count-1 do
        begin
          update_fields.GetItem(i,flddef);
          dsfield := ds.Fields[flddef.Index];
          params[i] := dsfield.AsVariant;
        end;
        UpdateRecords(update_sql, params);
      end;
      ds.Next;
    end;
  finally
    update_fields.free;
  end;
end;

procedure TJvRnnGenDBLink.ExecSQLStr(const sql : string);
var
  slist : TStrings;
begin
  slist := TStringList.Create;
  try
    slist.Text := sql;
    ExecuteStd(slist);
  finally
    slist.free;
  end;
end;

function TJvRnnGenDBLink.GetSQLFieldType(
      flddef : TFieldDef) : string;
      
  function sizeStr : string;
  begin
    Result := IntToStr(flddef.Size);
  end;

begin
  CASE flddef.DataType of
    ftString:
      Result := 'VARCHAR('+sizeStr+')';
    ftSmallInt:
      Result := 'SMALLINT';
    ftInteger:
      Result := 'INTEGER';
    ftFloat :
      Result := 'NUMERIC';
    ftDate :
      Result := 'DATE';
    ftTime :
      Result := 'TIME';
    ftDateTime :
      Result := 'TIMESTAMP';
    ftBoolean :
      Result := 'BOOLEAN';
  END;{CASE}
  Result := flddef.Name + ' '+Result;
end;

destructor TJvRnnGenDBLink.Destroy;
begin
  FDatasetLinks.CallActionMethod(FreeDatasetLink,nil);
  FDatasetLinks.Free;
  FConnectLink.Free;
  inherited;
end;

constructor TDatasetList.Create;
begin
  inherited CreateMgr(TJvCCObjRecMgr);
  KeyCompare := DSKeyCompare;
  ObjCompare := DSObjCompare;
  Sorted := TRUE;
end;


constructor TJvRnnDatasetLink.Create(dblink:TJvRnnGenDBLink);
begin
  FDBLink := dblink;
  inherited Create;
end;

destructor TJvRnnDatasetLink.Destroy;
begin
  if (FDBLink <> nil) then
    FDBLink.FDatasetLinks.DeleteItem(SELF);
  inherited;
end;

procedure TJvRnnDatasetLink.ForgetDataset;
begin
  FDataset := nil;
end;

constructor TDBConnectLink.Create;
begin
  FParams := TStringList.Create;
  inherited
end;

destructor TDBConnectLink.Destroy;
begin
  FParams.Free;
  inherited;
end;


procedure TDBConnectLink.SetParams(slist:TStrings);
begin
  FParams.Assign(slist);
end;



end.
