unit JvRnnProLs;

interface
uses TLHelp32,JvCCRecrd,JvCCArray;

type
  TJvRnnProcEntry = TProcessEntry32;
  PRnnProcEntry = ^TJvRnnProcEntry;
const
  kRnnMAXPROCENTRY = MAXINT div sizeof(TJvRnnProcEntry) - 1;
type
  TJvRnnProcEntryBuf = array[0..kRnnMAXPROCENTRY] of TJvRnnProcEntry;
  PRnnProcEntryBuf = ^TJvRnnProcEntryBuf;

  TJvRnnProcessListMgr = class(TJvCCRecordMgr)
    constructor Create; override;
    procedure ConstructItem(var buf); override;
  end;

  TJvRnnProcessList = class(TJvCCArray)
  private
    FPList : PRnnProcEntryBuf;
  public
    constructor Create(initsize:integer);
    function FindFullPath(const inname : string):PRnnProcEntry;
    function FindEXEName(const inname : string):PRnnProcEntry;

    property ArrayPtr : PRnnProcEntryBuf read FPList;
  end;

function GetRnnProcessList : TJvRnnProcessList;
function TerminateProcessID(inID : integer):boolean;

implementation
uses windows,sysutils,JvRnnStrUt;

function TerminateProcessID(inID : integer):boolean;
var
  hProcess : THandle;
begin
  hProcess := OpenProcess (PROCESS_TERMINATE,FALSE, inID);
  Result := (hProcess <> THandle(-1));
  if Result then
  try
    Result := TerminateProcess(hProcess, 0);
  finally
    CloseHandle (hProcess);
  end;
end;


constructor TJvRnnProcessListMgr.Create;
begin
  inherited Create;
  InitRecordMgr(sizeof(TJvRnnProcEntry),TRUE);
end;

procedure TJvRnnProcessListMgr.ConstructItem(var buf);
begin
  FillChar(buf,sizeof(TJvRnnProcEntry),#0);
end;

constructor TJvRnnProcessList.Create(initsize:integer);
begin
  inherited CreateMgr(TJvRnnProcessListMgr);
  SELF.BlockHdl := @FPList;
  Count := initsize;
end;

function TJvRnnProcessList.FindFullPath(const inname : string):PRnnProcEntry;
var
  i : integer;
  p : PRnnProcEntry;
begin
  Result := nil;
  for i := 0 to Count-1 do
  begin
    p := Ptr[i];
    if COMPARETEXT(inname,p^.szExeFile)=0 then
    begin
      Result := p;
      exit;
    end;
  end;
end;

function TJvRnnProcessList.FindEXEName(const inname : string):PRnnProcEntry;
var
  i : integer;
  p : PRnnProcEntry;

  fname : string;
  prefix : string;
  ext : string;
begin
  Result := nil;
  for i := 0 to Count-1 do
  begin
    p := Ptr[i];
    fname := ExtractFileName(p^.szExeFile);
    ExtractFileNameExt(fname,prefix,ext);
    if (COMPARETEXT(inname,prefix)=0)
      OR (COMPARETEXT(inname,fname)=0) then
    begin
      Result := p;
      exit;
    end;
  end;
end;
(*
function GetProcessModule(dwPID, dwModuleID:DWORD;
  var pme32 : TModuleEntry32) : BOOL;
var
  bFound : BOOL;
  hModuleSnap : THandle;
  me32 : TModuleEntry32;
begin
    // Take a snapshot of all modules in the specified process.
  Result := FALSE;
  hModuleSnap := CreateToolhelp32Snapshot(TH32CS_SNAPMODULE, dwPID);
  if (hModuleSnap = THandle(-1)) then
    exit;

  try
    // Fill the size of the structure before using it.
    me32.dwSize := sizeof(TModuleEntry32);

      // Walk the module list of the process, and find the module of
      // interest. Then copy the information to the buffer pointed
      // to by lpMe32 so that it can be returned to the caller.
    if (Module32First(hModuleSnap, me32)) then
    repeat
      if (me32.th32ModuleID = dwModuleID) then
      begin
        pme32 := me32;
        Result := TRUE;
      end;
    until( bFound OR (not Module32Next(hModuleSnap, me32)));
  finally
    // Do not forget to clean up the snapshot object.
    CloseHandle (hModuleSnap);
  end;
end;
*)
function GetRnnProcessList : TJvRnnProcessList;
var
  hSnapshot : THandle;
  pe32 : TJvRnnProcEntry;
  hProcess : THandle;
begin
  Result := nil;
  hSnapshot := CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0);
  if (hSnapshot = THandle(-1)) then
    exit;
  try
    Result := TJvRnnProcessList.Create(0);

    pe32.dwSize := sizeof(TJvRnnProcEntry);
    if Process32First(hSnapshot, pe32) then
    repeat
      hProcess := OpenProcess (PROCESS_TERMINATE,FALSE, pe32.th32ProcessID);
      if (hProcess <> THandle(-1)) then
      try
        Result.InsertInto(Result.Count,pe32);
      finally
        CloseHandle (hProcess);
      end;
    until not Process32Next(hSnapshot, pe32);
  finally
    CloseHandle (hSnapshot);
  end;
end;

end.


