{-----------------------------------------------------------------------------
The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in compliance
with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/MPL-1.1.html

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either expressed or implied. See the License for
the specific language governing rights and limitations under the License.

The Original Code is: JvRnnProfl.PAS, released on 2000-12-01.

The Initial Developer of the Original Code is Michael Rynn.
Portions created by Michael Rynn are Copyright (C) 2000  Michael Rynn.
All Rights Reserved.

Contributor(s): ______________________________________.

Last Modified: 2000-12-01

You may retrieve the latest version of this file at the Project JEDI home page,
located at http://www.delphi-jedi.org

-----------------------------------------------------------------------------}
{$A+,B-,C+,D+,E-,F-,G+,H+,I+,J+,K-,L+,M-,N+,O+,P+,Q-,R-,S-,T-,U-,V+,W-,X+,Y+,Z1}

{$I JCL.INC}
unit JvRnnProfl;
interface
Uses SysUtils,Classes,JvRnnLogTx;

const
  prof_logresource = 1;
  prof_exitdump = 2;
  prof_all = prof_logresource + prof_exitdump;
  {changeable}
  gLogStamps : Boolean = TRUE;
  gLogExceptions : Boolean = TRUE;
  gProfileTimeRatio : double = 1.0;
type
  ERnnProfile = class(Exception)
    constructor Create(const inStr : string);
  end;

  TRnnProfileOptions = integer;
  TSegNameString = string[63];

procedure StampEntry(const stampName : TSegNameString); stdcall;
procedure StampExit(options:TRnnProfileOptions);
procedure StampException(options:TRnnProfileOptions); stdcall;
procedure DumpResources(const atName : string);
procedure DumpProfileLog;
procedure ResetAllProfileLogs;
Function DelphiHeapStatus : string;
Function WinMemStatus : string;

var
  gLogFileName : string;
  gRnnProfLog : TRnnLogTextFile;

implementation
uses
  WinProcs,Wintypes,
  Forms, JvRnnStrUt, JvCCRecrd, JvCCArray;

const
  kInitProfStack = 32;
  kInitProfSegments = 50;
  kMaxProfileStack = 500; {recursion must be constrained!}

type
  PRnnProfSegment = ^TRnnProfSegment;
  TRnnProfSegment = record
    FSegName : TSegNameString;
    FSegSumTime : double;
    FSegSumSq :  double;
    FSegMaxTime : double;
    FSegMinTime : double;
    FSegEntries : LongInt;
    FRecursion : integer;
  end;


const
  kRnnProfSegArray = (MAXINT div sizeof(TRnnProfSegment))  - 1;
type
  TRnnProfSegArray = array[0..kRnnProfSegArray] of TRnnProfSegment;
  PRnnProfSegArray = ^TRnnProfSegArray;

  TRnnPSegArray = array[0..kRnnProfSegArray] of PRnnProfSegment;
  PRnnPSegArray = ^TRnnPSegArray;

  TRnnProfSegments = class;

  TRnnProfSegPtrs = class(TJvCCIntArray)
  public
    constructor Create(initCapacity : integer);
  end;

  TRnnProfSegMgr = class(TJvCCRecordMgr)
    constructor Create; override;
    procedure SwapItems(var p1, p2); override;
  end;

  TRnnProfSegments = class(TJvCCArray)
  protected
    FProfArray : PRnnProfSegArray;
    FSegPtrs : TRnnProfSegPtrs;
  public
    constructor Create(initCapacity : integer);
    destructor Destroy; override;
    function FindSegPtr(const s : TSegNameString; var pix : integer) : integer;
    function NewSegment(const s :  TSegNameString; pix : integer) : integer;
    property ArrayPtr : PRnnProfSegArray read FProfArray write FProfArray;
  end;

  TRnnProfTimeRec = record
    entrytime : TDateTime;
    exittime : TDateTime;
    profileix : integer;
  end;

  TRnnProfTimeMgr = class(TJvCCRecordMgr)
    constructor Create; override;
    procedure SwapItems(var p1, p2); override;
  end;

const
  kRnnProfTimeArray = (MAXINT div sizeof(TRnnProfTimeRec))  - 1;
type
  TRnnProfTimeArray = array[0..kRnnProfTimeArray] of TRnnProfTimeRec;
  PRnnProfTimeArray  = ^TRnnProfTimeArray;

  TRnnProfStack = class(TJvCCArray)
  protected
    FStampArray : PRnnProfTimeArray;
  public
    constructor Create(initCapacity : integer);
    property ArrayPtr : PRnnProfTimeArray read FStampArray write FStampArray;
  end;
var
  gPrSegments : TRnnProfSegments;
  gPrStack : TRnnProfStack;

procedure RecordEntry(
  var prof : TRnnProfSegment; inTime : TDateTime);forward;
procedure RecordExit(
  var prof : TRnnProfSegment; inseconds : double);forward;
procedure GetProfileStats(
  const prof : TRnnProfSegment;
  var entryCount : integer;
  var sum, minTime, maxTime, aveTime, stdevTime:double);forward;
procedure ResetProfile(
  var prof : TRnnProfSegment); forward;
function SecondsToStr(
  inTime:double):string; forward;

procedure StampEntry(const stampName : TSegNameString);
var
  profix : integer;
  ix : integer;
  stackct : integer;
begin
  if gPrSegments = nil then
  begin
    gPrSegments := TRnnProfSegments.Create(kInitProfSegments);
    gPrStack := TRnnProfStack.Create(kInitProfStack);
    if gRnnProfLog=nil then
    begin
      if Length(gLogFileName)=0 then
      begin
        gLogFileName := ExtractFilePath(Application.ExeName);
        RemoveEndChar(gLogFileName, '\');
        gLogFileName := gLogFileName + '\ProflOut.txt';
      end;
      gRnnProfLog := TRnnLogTextFile.Create(gLogFileName);
      gRnnProfLog.OpenAppend;
      gRnnProfLog.Append('--------------- Profile Log Start ---------------');
    end;
  end;
  if gPrStack.Count > kMaxProfileStack then
    raise ERnnProfile.Create('RnnProfiler stack overflow');
  profix := gPrSegments.FindSegPtr(stampName, ix);
  if profix < 0 then
    profix := gPrSegments.NewSegment(stampName, ix);
  stackct :=  gPrStack.Count;
  gPrStack.Count := stackct+1;
  with gPrStack.ArrayPtr^[stackct] do
  begin
    profileix := profix;
    entryTime := Now;
    exitTime := 0;
    RecordEntry(gPrSegments.ArrayPtr^[profileix],entryTime);
  end;
end;

function SecondsToStr(inTime:double):string;
{return time in decimal seconds}
begin
  Result := Format('%10.4f',[inTime]);
end;


Function PopProfileStack(var timeTaken : double) : integer;
var
  tempTime : TDateTime;
  stackct : integer;
begin
  tempTime := Now;
  stackct := gPrStack.Count-1;
  if (stackct >= 0) then
  begin
    WITH gPrStack.ArrayPtr^[stackct] do
    begin
      exitTime := tempTime;
      timeTaken := (exitTime - entryTime)*24.0*60*60;
      RecordExit(gPrSegments.ArrayPtr^[profileix],timeTaken);
      Result := profileix;
    end;
    gPrStack.Count := stackct;
  end else
    result := -1;
end;

procedure StampException(options : TRnnProfileOptions);
var
  timeTaken : double;
  prof : PRnnProfSegment;
  profix : integer;
  eobj : Tobject;
  e : Exception;
  s  : string;
begin
  profix := PopProfileStack(timeTaken);
  if profix >= 0 then
    prof := gPrSegments.Ptr[profix]
  else
    prof := nil;
  eobj := ExceptObject;
  if (eobj <> nil) then
  begin
    e := Exception(eobj);
    if gLogExceptions and (prof <> nil) and (gRnnProfLog<>nil) then
    begin
      s := Format( '%3d:'#9'%40s'#9'%s'#9'%s',
        [gPrStack.Count,prof^.FSegName,
         SecondsToStr(timeTaken),e.Message]);
      gRnnProfLog.Append(s);
    end;
  end;
end;

procedure StampExit(options : TRnnProfileOptions);
var
  timeTaken : double;
  prof : PRnnProfSegment;
  profix : integer;
  s : string;
begin
  profix := PopProfileStack(timeTaken);
  if profix >= 0 then
    prof := @gPrSegments.ArrayPtr^[profix]
  else
    prof := nil;
  if (prof = nil) then
    raise ERnnProfile.Create('StampExit: Profile stack empty');
  if gRnnProfLog<>nil then
  begin
    if gLogStamps and ((prof_logresource and options)<>0) then
    begin
      s := Format('%3d:'#9'%40s'#9'%s'#9'%s'#9'%s',
        [gPrStack.Count,prof^.FSegName,SecondsToStr(timeTaken),
         DelphiHeapStatus,WinMemStatus]);
      gRnnProfLog.Append(s);
    end;
    if ((prof_exitdump and options)<>0) then
      DumpProfileLog;
  end;
end;

procedure DumpResources(const atName : string);
var
  s : string;
begin
  if gRnnProfLog = nil then
    exit;
  s := Format('%s'#9'%s'#9'%s',
    [atname, DelphiHeapStatus,WinMemStatus]);
  gRnnProfLog.Append(s);
end;

Function WinMemStatus : string;
begin
{$IFNDEF WIN32}
  Result := Format('Win heap %d bytes. System %d GDI %d. User %d (%%free)',
    [GetFreeSpace(0),
     GetFreeSystemResources(GFSR_SYSTEMRESOURCES),
     GetFreeSystemResources(GFSR_GDIRESOURCES),
     GetFreeSystemResources(GFSR_USERRESOURCES)]);
{$ELSE}
  Result := '';
{$ENDIF}
end;


Function DelphiHeapStatus : string;
begin
{$IFNDEF WIN32}
  Result := ' ' (*Format('alloc %d free %d segs %d',
				[ckbytesAllocated, ckFreeBytes, ckSegsAllocated]); *)
{$ELSE}
  Result := Format('%d blocks, %d bytes',
				[AllocMemCount, AllocMemSize]);
{$ENDIF}

end;



constructor ERnnProfile.Create(const inStr : string);
begin
  if gRnnProfLog<>nil then
  begin
    gRnnProfLog.Append(inStr);
    gRnnProfLog.Free;
    gRnnProfLog := nil;
  end;
  inherited Create(inStr);
end;


procedure RecordEntry(var prof : TRnnProfSegment; inTime : TDateTime);
begin
  Inc(prof.FRecursion);
end;

procedure RecordExit(var prof  : TRnnProfSegment; inSeconds : double);
begin
  WITH prof do
  BEGIN
    Dec(FRecursion);
    if FRecursion = 0 then
    begin
      Inc(FSegEntries);
      if FSegEntries = 1 then
      begin
        FSegMinTime := inSeconds;
        FSegMaxTime := inSeconds;
      end else
      begin
        if (inSeconds < FSegMinTime) then
          FSegMinTime := inSeconds
        else if (inSeconds > FSegMaxTime) then
          FSegMaxTime := inSeconds
      end;
      FSegSumTime := FSegSumTime + inSeconds;
      FSegSumSq := FSegSumSq + inSeconds*inSeconds;
    end;
  END;
end;

procedure GetProfileStats(const prof : TRnnProfSegment; var entryCount : integer;
                  var sum, minTime, maxTime, aveTime, stdevTime:double);
begin
  WITH prof do
  BEGIN
    entryCount := fSegEntries;
    minTime := FSegMinTime;
    maxTime := FSegMaxTime;
    sum := FSegSumTime;
    if entryCount > 1 then
    begin
      aveTime := FSegSumTime / entryCount;
      stdevTime := (FSegSumSq - entryCount*aveTime*aveTime) / (entryCount-1);
    end else
    begin
      aveTime := FSegSumTime;
      stdevTime := 0;
    end;
  END;
end;

procedure ResetProfile(var prof : TRnnProfSegment);
begin
  WITH prof do
  BEGIN
    FSegSumTime := 0;
    FSegMaxTime := 0;
    FSegMinTime := 0;
    FSegSumSq := 0;
    FSegEntries := 0;
    FRecursion := 0;
  END;
end;


procedure ResetAllProfileLogs;
var
  i : integer;
  prof : PRnnProfSegment;
begin
  prof := @gPrSegments.ArrayPtr[0];
  for i := 0 to gPrSegments.Count-1 do
  begin
    ResetProfile(prof^);
    Inc(prof);
  end;
end;

procedure DumpProfileLog;
var
  i : integer;
  prof : PRnnProfSegment;
  entryCount : integer;
  sum, min, max, ave, stdev : double;
  s : string;
begin
  if gRnnProfLog = nil then
    exit;
  gRnnProfLog.Append('Profile dump at '+DateTimeToStr(Now));
  s := Format('%-40s %7s: %10s %7s %7s %7s %7s',
         ['Profile name', 'entries', 'total',
           'min', 'max', 'avg', 'stdev (s)']);
  gRnnProfLog.Append(s);
  prof := @gPrSegments.ArrayPtr[0];
  for i := 0 to gPrSegments.Count-1 do
  begin
    GetProfileStats(prof^,entryCount,sum,  min, max, ave, stdev);
    sum := sum *gProfileTimeRatio;
    min := min *gProfileTimeRatio;
    max := max *gProfileTimeRatio;
    ave := ave *gProfileTimeRatio;
    stdev := stdev *gProfileTimeRatio;
    s := Format('%-40s %7d: %10.4f %7.4f %7.4f %7.4f %7.4f',
      [prof^.FSegName, entryCount, sum, min, max, ave, stdev]);
    gRnnProfLog.Append(s);
    Inc(prof);
  end;
  gRnnProfLog.Append('');
end;

constructor TRnnProfSegMgr.create;
begin
  inherited create;
  InitRecordMgr(sizeof(TRnnProfSegment),TRUE);
end;

procedure TRnnProfSegMgr.SwapItems(var p1, p2);
var
  seg : TRnnProfSegment;
begin
  seg := TRnnProfSegment(p1);
  TRnnProfSegment(p1) := TRnnProfSegment(p2);
  TRnnProfSegment(p2) := seg;
end;

constructor TRnnProfSegments.Create(initCapacity : integer);
begin
  inherited CreateMgr(TRnnProfSegMgr);
  BlockHdl := @FProfArray;
  Capacity := initCapacity;
  FSegPtrs := TRnnProfSegPtrs.Create(initCapacity);
end;

destructor  TRnnProfSegments.Destroy;
begin
  FSegPtrs.Free;
  inherited Destroy;
end;


function TRnnProfSegments.FindSegPtr(const s : TSegNameString; var pix : integer) : integer;
var
  pseg : PRnnProfSegment;
  ix : integer;
  iar : PCCIntBuf;
  cpres : integer;
  mid : integer;
  L, H : integer;
begin
  {Result is index to actual TRnnProfSegment,
   pix is index of FSegPtrs}
  iar := FSegPtrs.BlockPtr;
  L := 0;
  H := FSegPtrs.Count-1;
  if L <= H then
  repeat
    mid := ((H - L) shr 1) + L;
    ix := iar^[mid];
    pseg := Ptr[ix];
    cpres := COMPARESTR(pseg^.FSegName,s);
    if (cpres > 0) then {spot is below mid}
      H := mid-1
    else if (cpres < 0) then {spot is above mid}
      L := mid+1
    else
    begin
      Result := ix;
      exit;
    end;
  until(L>H);
  pix := L; {insert index for FSegPtrs}
  Result := -1;
end;

function	TRnnProfSegments.NewSegment(
  const s :  TSegNameString; pix : integer) : integer;
var
  p : PRnnProfSegment;
begin
 {make new space, pix is the insertion index for FSegPtrs}
  Result := Count; {position of new record}
  FSegPtrs.CreateItemsAt(pix,1); {insert a new index at position pix}
  Count := Result + 1;   {increase the array of records}
  p := Ptr[Result]; {point to the new record}
  p^.FSegName := s; {initialize name}
  FSegPtrs.PArray^[pix] := Result; {set the index}
end;

constructor TRnnProfTimeMgr.create;
begin
  inherited create;
  InitRecordMgr(sizeof(TRnnProfTimeRec),FALSE);
end;

procedure TRnnProfTimeMgr.SwapItems(var p1, p2);
var
  rec : TRnnProfTimeRec;
begin
  rec := TRnnProfTimeRec(p1);
  TRnnProfTimeRec(p1) := TRnnProfTimeRec(p2);
  TRnnProfTimeRec(p2) := rec;
end;

constructor TRnnProfStack.Create(initCapacity : integer);
begin
  inherited CreateMgr(TRnnProfTimeMgr);
  BlockHdl := @FStampArray;
  Capacity := initCapacity;
end;

constructor TRnnProfSegPtrs.Create(initCapacity : integer);
begin
  inherited Create(-1);
  Capacity := initCapacity;
end;

end.
