{-----------------------------------------------------------------------------
The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in compliance
with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/MPL-1.1.html

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either expressed or implied. See the License for
the specific language governing rights and limitations under the License.

The Original Code is: JvRnnScBDE.PAS, released on 2000-12-01.

The Initial Developer of the Original Code is Michael Rynn.
Portions created by Michael Rynn are Copyright (C) 2000  Michael Rynn.
All Rights Reserved.

Contributor(s): ______________________________________.

Last Modified: 2000-12-01

You may retrieve the latest version of this file at the Project JEDI home page,
located at http://www.delphi-jedi.org

-----------------------------------------------------------------------------}
{$A+,B-,C+,D+,E-,F-,G+,H+,I+,J+,K-,L+,M-,N+,O+,P+,Q-,R-,S-,T-,U-,V+,W-,X+,Y+,Z1}

{$I JCL.INC}
unit JvRnnScBDE;

interface
Uses Classes,DB,DBTables,SysUtils,JvCCArray,JvRnnScrSt,JvRnnScCst,JvRnnVarCp;

type

  TJvRnnBDEScript = class;

  ERnnBDEScript = class(ERnnScript);

  TJvRnnTableNotifyEvent =  procedure(Sender:TObject;
    vdb : TDatabase; const s : string) of object;


  TJvRnnBDEScript = class(TJvRnnCustomScript)
  private
    FDatabase : TDatabase;
    FAlias : string;
    FAliasDB : TDatabase;  (* cache database with name FAlias *)
    FPrivDB : TDatabase;
    FQuery : TQuery;

    FViewTableName : string;
    FBeforeViewCreate : TJvRnnTableNotifyEvent;
    FAfterViewCreate : TJvRnnTableNotifyEvent;
    FBeforeViewDestroy : TJvRnnTableNotifyEvent;

    procedure SetPrivateDB(inDB : TDatabase);
    procedure SetDatabase(inDB : TDatabase);

  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure ExecuteStd(cmd : TStrings); override;
  public
    constructor Create(AOwner:TComponent); override;
    destructor Destroy;override;

    {script processing}
    procedure SetAliasName(const aliasname:string);
    function  GetAliasedDB(const inAlias:string):TDatabase;

    function ConnectAlias(const aliasname : string;  params : TStrings) : boolean;
    property Alias : string read FAlias write SetAliasName;
    property ViewTableName : string read FViewTableName write FViewTableName;
    function IsOracleDB(db : TDatabase) : boolean;
    function ExecQuery(const dbname : string; const qrylabel : string) : TDataset;
  published
    property BeforeCmdExecute;

    property DateFormat;
    property DateTimeFormat;
    property DeleteComments;
    property ParamChar;
    property PrefixChar;
    property ScriptStore;
    property Terminator;
    property TimeFormat;
    property ValueList;

    property Database : TDatabase read FDatabase write SetDatabase;
    property PrivateDB : TDatabase read FPrivDB write SetPrivateDB;
    property BeforeViewCreate : TJvRnnTableNotifyEvent read FBeforeViewCreate write FBeforeViewCreate;
    property AfterViewCreate: TJvRnnTableNotifyEvent read FAfterViewCreate write FAfterViewCreate;
    property BeforeViewDestroy: TJvRnnTableNotifyEvent read FBeforeViewDestroy write FBeforeViewDestroy;
    property OnLogStrings;
  end;


implementation
Uses Forms,JvRnnStrUt,JvRnnUtBDE,Dialogs,Controls,JvRnnPars,JvRnnStrLs
 {$ifdef WIN32},JvRnnCroTb {$endif};

{$IFNDEF RL_D3}
type
  TBDEDataset = TDataset;
{$ENDIF}

{built in procedures for TJvRnnBDEScript}
const
  kCreateQry_F = 'CREATE_QRY';
  kDeleteQry_F = 'DELETE_QRY';
  kUpdateQry_F = 'UPDATE_QRY';
  kInsertQry_F = 'INSERT_QRY';
  kDeleteTable_F = 'DROP_TABLE';
  kReadTable_F = 'READ_TABLE';

  kAbortOnZero_F = 'ABORT_ON_ZERO';
  kDisplay_F	= 'DISPLAY';
  kInsertView_F = 'INSERT_QRY';
  kUpdateView_F = 'UPDATE_QRY';

  kForRows_F = 'FOR_ROWS';
  kEndForRows_F = 'END_FOR_ROWS';
  

  kBDEAddAlias_F = 'CREATE_ALIAS';
  kBDEModifyAlias_F = 'MODIFY_ALIAS';
  kConnectDatabase = 'CONNECT_ALIAS';

  kBDETableExists = 'BDE_TABLE_EXISTS';
  kBDEAliasExists = 'BDE_ALIAS_EXISTS';

	(*kSet_F = 'SET';
	kAlias_F = 'ALIAS';
	*)

{-------------- functions -----------------------}
function BDETableExists_F(sct: TJvRnnCustomScript; args : TStrings): string; far; forward;
{$IFDEF RL_D3}
function BDEAliasExists_F(sct:TJvRnnCustomScript; args : TStrings): string; far; forward;

procedure BDEAddAlias_F(data:Pointer); far; forward;
{@CREATE_ALIAS <Alias Name> <Driver Name> <parameters list>}

procedure BDEModifyAlias_F(data:Pointer); far; forward;
{@MODIFY_ALIAS <Alias Name> <Driver Name> <parameters list>}
{$ENDIF}

{-------------- View / Table functions ----------------}

procedure CreateView_F(data:Pointer); far; forward;
{@CREATE_VIEW ':ALIAS:View.db' [primary key count] as select ... (sql)}
procedure DeleteView_F(data:Pointer); far; forward;
{@DELETE_QRY ':ALIAS:View.db' }
{$ifdef win32}
procedure ReadView_F(data:Pointer); far; forward;
{$endif}

{@READ_TABLE '[:ALIAS:]View.db' ASSIGN(<AssignList>)
    LOCATE(FieldName1=Value1 [,FieldName=Value])}
{@READ_QRY '[:ALIAS:]View.db' (<AssignList>) [LOCATE(...)] select ... }

{<AssignList> -> Varname=FieldName [,Varname=FieldName]}

procedure Display_F(data:Pointer); far; forward;
{@VIEW ':ALIAS:View.db' }
procedure Query_F(data:Pointer); far; forward;
{@QUERY select ...(sql) }
procedure BatchView_F(data:Pointer); far; forward;
{@INSERT_QRY ':ALIAS:View.db' [USE_FIELDNAMES] select ... (sql) }
procedure GetRecordCount_F(data:Pointer); far; forward;
{@GET_RECORD_COUNT <variable name> tablename;}

procedure SetAlias_F(data:Pointer); far; forward;
{@SET_ALIAS <BDE Alias Name> <BDE Driver Name> }


procedure ForRows_F(data:Pointer);far; forward;
(*
procedure Set_F(data:Pointer); far; forward;
{@SET [variable name] [value]}
procedure Alias_F(data:Pointer); far; forward;
{@ALIAS [aliasname]}
*)

procedure EndForRows_F(data:Pointer); far; forward;

procedure AbortOnZero_F(data:Pointer); far; forward;
{@ABORT_ON_ZERO variable msg}


procedure BDEConnectAlias_F(data:Pointer); far; forward;
{@MODIFY_ALIAS <Alias Name> <parameters list>}

constructor TJvRnnBDEScript.Create(AOwner:TComponent);
begin
 	inherited create(AOwner);

  DateFormat := 'dd-mmm-yyyy hh:nn:ss';
  
  FQuery := TQuery.Create(SELF);

  InstallCustomScriptOps(SELF);

  SetUserOp('CREATE_QRY',@CREATEVIEW_F);
  SetUserOp('QUERY',@QUERY_F);
  SetUserOp('DROP_TABLE',@DeleteView_F);
  SetUserOp('ABORT_ON_ZERO',@AbortOnZero_F);
  SetUserOp('DISPLAY',@Display_F);
  {$ifdef WIN32}
  SetUserOp(kReadTable_F, @ReadView_F);
  SetUserOp('READ_QUERY', @ReadView_F);
  InstallBDECrossTab(SELF);
  {$endif}
  SetUserOp('INSERT_QRY',@BatchView_F);
  SetUserOp('DELETE_QRY',@BatchView_F);
  SetUserOp('UPDATE_QRY',@BatchView_F);
  SetUserOp(kForRows_F,@ForRows_F);
  SetUserOp(kEndForRows_F, @EndForRows_F);

  SetUserOp('SET_ALIAS', @SetAlias_F);
  SetUserOp('GET_RECORD_COUNT', @GetRecordCount_F);
  SetUserOp(kConnectDatabase,@BDEConnectAlias_F);

  SetUserFunc(kBDETableExists, @BDETableExists_F);
  {$IFDEF RL_D3}
  SetUserOp(kBDEAddAlias_F, @BDEAddAlias_F);
  SetUserOp(kBDEModifyAlias_F, @BDEModifyAlias_F);
  SetUserFunc(kBDEAliasExists, @BDEAliasExists_F);
  {$ENDIF}

end;

destructor TJvRnnBDEScript.Destroy;
begin
  FUserOps.Free;
  FUserFuncs.Free;
  inherited Destroy;
end;

function TJvRnnBDEScript.IsOracleDB(db : TDatabase) : boolean;
{$IFDEF VER80}
var
  plist : TStrings;
{$ENDIF}
begin
{$IFDEF VER80}
  plist := TStringList.create;
  TRY
    Session.GetAliasParams(Database.AliasName, plist);
    Result := plist.Values['TYPE'] = 'ORACLE';
  FINALLY
    plist.free;
  END;
{$ELSE}
  Result := (db.Session.GetAliasDriverName(db.DatabaseName) = 'ORACLE');
{$ENDIF}
end;


function TJvRnnBDEScript.ExecQuery(
  const dbname : string; const qrylabel : string) : TDataset;
var
  qry : TQuery;
  slist : TStrings;
begin
  CheckHasStore;
  slist := TStringlist.create;
  try
    Result := nil;
    ScriptStore.GetScriptAsList(qrylabel,slist);
    DefaultCmdProcess(slist);
    qry := TQuery.Create(SELF);
    qry.databasename := dbname;
    qry.SQL.Assign(slist);
    try
      qry.Active := TRUE;
      Result := qry;
    except
      qry.Free;
      raise;
    end;
  finally
    slist.free;
  end;
end;

procedure TJvRnnBDEScript.SetDatabase(inDB : TDatabase);
begin
  FDatabase := inDB;
end;

procedure TJvRnnBDEScript.SetPrivateDB(inDB : TDatabase);
begin
  FPrivDB := inDB;
end;

function TJvRnnBDEScript.ConnectAlias(const aliasname : string; params : TStrings) : boolean;
var
  db : TDatabase;
  aliasParams : TJvRnnStrList;
  prompt : boolean;
  i : integer;
  s : string;
  copyalias : string; (* something buggy with string *)
begin
  aliasParams := TJvRnnStrList.Create;
  TRY
    copyalias := aliasname;
	   Session.GetAliasParams(copyalias, aliasParams);
    prompt := FALSE;
    db := Session.FindDatabase(copyalias);
    if (db = nil) then
    begin
      db := TDatabase.Create(SELF);
      WITH db do begin
        DatabaseName := copyalias;
        s := GetDBDriverName(db);
        DriverName := s;
      end;
    end;

	   REPEAT {until connected}
      aliasParams.AddStrings(params);
      for i := 0 to params.Count-1 do
      begin
        s := GetStrName(params[i]);
        aliasParams.Values[s] := params.Values[s];
      end;
      WITH db do
      begin
        LoginPrompt := prompt;
        Params.Clear;
        Params.Assign(aliasParams);
        TRY
          Connected := TRUE;
        EXCEPT
          On ex2:EDatabaseError do
          begin
            if (MessageDlg('Connect failed: '+ex2.Message+' Try again?',
                  mtConfirmation,[mbYes,mbNo],0)<>mrYes) then
              raise;
            prompt := TRUE;
          end;
        END;
	     end;
    UNTIL(db.Connected);
  FINALLY
    aliasParams.Free;
  END;
  Result := db.Connected;
end;


procedure TJvRnnBDEScript.Notification(AComponent: TComponent; Operation: TOperation);
begin
  if Operation = opRemove then begin
    if AComponent = Database then
      Database := nil;
  end;
  inherited Notification(AComponent, Operation);
end;


procedure GetRecordCount_F(data:Pointer);
var
  cmdInfo : PJvScriptCmdInfo;
  p : PChar;
  scriptor : TJvRnnBDEScript;
  token : string;
  tablename, aliasname : string;
  varname : string;
  ptype : TJvRnnParenType;
  tokentype : TJvRnnTokenType;
  reccount : integer;
  db : TDatabase;

  function sub_next : boolean;
  begin
    Result := StrNextTokenType(p,token,ptype,tokentype);
  end;

begin
  cmdInfo := PJvScriptCmdInfo(data);
  scriptor := cmdInfo^.scriptor as TJvRnnBDEScript;

  p := cmdInfo^.cmdcstr;

  if ( not sub_next) then
    ERnnBDEScript.create('variable name expected');

  varname := token;

  if (not sub_next) then
    ERnnScript.create('table name expected');

  if GetAliasTableName(token,aliasname, tablename) then
  begin
    if Length(aliasname)>0 then
      db := scriptor.GetAliasedDB(aliasname)
    else
      db := scriptor.FAliasDB;

    reccount := DBTableRowCount(db,tablename);
    scriptor.SetVariable(varname,IntToStr(reccount));
  end;
end;



procedure SetAlias_F(data:Pointer);
var
  cmdInfo : PJvScriptCmdInfo;
  p : PChar;
  scriptor : TJvRnnBDEScript;
  token : string;
  ptype : TJvRnnParenType;

  procedure sub_next;
  begin
    if not StrNextToken(p,token,ptype) then
      raise ERnnBDEScript.create('No aliasname');
  end;
begin

	cmdInfo := PJvScriptCmdInfo(data);
	scriptor := cmdInfo^.scriptor as TJvRnnBDEScript;

	p := cmdInfo^.cmdcstr;

	sub_next;

	scriptor.Alias := token;
end;


type
  TForRowsObj = class(TJvRnnScrObj)
  protected
    FFieldNames : TStrings;
    FSavedVar : TStrings;
    FLoopBackLine : integer;
    FLoopTable : TTable;
    FFirstPass : boolean;
    FScriptor : TJvRnnBDEScript;

    function InitLoop(cmdInfo : PJvScriptCmdInfo) : boolean;
    function NextRow(var ss : TJvRnnScriptState) : boolean;
    procedure AfterPopState(var rst:TJvRnnScriptState); override;
    procedure RestoreVariables;
    destructor Destroy; override;
  end;

procedure TForRowsObj.AfterPopState(var rst:TJvRnnScriptState);
begin
  RestoreVariables;
end;


destructor TForRowsObj.Destroy;
begin
  FLoopTable.Free;
  FFieldNames.Free;
  FSavedVar.Free;

  inherited Destroy;
end;

procedure TForRowsObj.RestoreVariables;
var
  varray : TJvCCStrStrArray;
  i : integer;
  vix : integer;
begin
  if (not FFirstPass) and (FFieldNames <> nil) then
  begin
    varray := FScriptor.Variables;

    for i := 0 to FFieldNames.count-1 do
    begin
    {remove temporary variables with same name}
      vix := varray.IndexOf(FFieldNames[i]);
      if vix >= 0 then
        varray.DeleteItemsAt(vix,1);
    end;
  {restore variables which had same name}
    if FSavedVar <> nil then
      varray.AppendValueStrings(FSavedVar);
  end;
end;

function TForRowsObj.NextRow(var ss : TJvRnnScriptState) : boolean;
var
  varray : TJvCCStrStrArray;
  i : integer;
  vix : integer;
begin
  if FFirstPass then
    FLoopTable.First
  else
    FLoopTable.Next;

  if FLoopTable.EOF then
  begin
    Result := FALSE;
    exit;
  end;

  if (FFieldNames<> nil) then
  begin
    varray := FScriptor.Variables;
    for i := 0 to FFieldNames.count-1 do
    begin
    {SAVE any global variables with same name}
      if FFirstPass then
      begin
        vix := varray.IndexOf(FFieldNames[i]);
        if vix >= 0 then
        begin
          if (FSavedVar=nil) then
            FSavedVar := TJvRnnStrList.create;
          FSavedVar.Values[FFieldNames[i]] := varray.GetIndexValue(vix);
        end;
      end;
      varray.SetVariable(FFieldNames[i],TField(FFieldNames.Objects[i]).DisplayText);
    end;
  end;
  ss.lineNum := FLoopBackLine;
  FFirstPass := FALSE;
  Result := TRUE;

end;

function TForRowsObj.InitLoop(cmdInfo : PJvScriptCmdInfo):boolean;
const
  kForSyntax = '@FOR_ROWS [| * | name1,name2..| ] in ''Table''';
var
  p, r : PChar;
  token : string;
  xtoken : string;
  ptype : TJvRnnParenType;
  tokentype : TJvRnnTokenType;
  allfields : boolean;
  aliasname : string;
  tblname : string;
  db : TDatabase;
  i : integer;
  anyField : TField;
  dtfield : TDateTimeField;
  ss : TJvRnnScriptState;
  dct : integer;

  procedure sub_next;
  begin
    if not StrNextTokenType(p,token,ptype,tokentype) then
      raise ERnnBDEScript.create(kForSyntax);
  end;
begin
  Result := FALSE;
  FScriptor := cmdInfo^.scriptor as TJvRnnBDEScript;
  p := cmdInfo^.cmdcstr;
  sub_next;

  if not (COMPARETEXT('in',token)=0) then
  begin
    FFieldNames := TStringList.Create;
    allfields := (COMPARETEXT('*',token)=0);
    if not allfields then
    begin
      while (COMPARETEXT(token,'in')<>0) do
      begin
        if token <> ',' then
          FFieldNames.Add(token);
        sub_next;
      end;
    end else
    begin
      sub_next;
    end;
    if not (COMPARETEXT(token,'in')=0) then
      raise ERnnBDEScript.create(kForSyntax);
  end else
  begin
    allfields := FALSE;
  end;
  {get table name}
  sub_next;

  GetAliasTableName(token,aliasname,tblName);

  if Length(aliasName) > 0 then
    db := FScriptor.GetAliasedDB(aliasName)
  else
    db := FScriptor.FAliasDB;

  FLoopTable := DBOpenTable(db,tblName);

  if allfields then
  begin
    for i := 0 to FLoopTable.Fieldcount-1 do
      FFieldNames.Add(FLoopTable.Fields[i].FieldName);
  end;
  for i := 0 to FFieldNames.count-1 do
  begin
    anyField := FLoopTable.FieldByName(FFieldNames[i]);
    if (Length(FScriptor.DateFormat)>0)
      and (anyField is TDateTimeField) then
    begin
      dtField := TDateTimeField(anyField);
      CASE (dtField.DataType) of
      ftDate : dtField.DisplayFormat := FScriptor.DateFormat;
      ftTime :  dtField.DisplayFormat := FScriptor.TimeFormat;
      ftDateTime : dtField.DisplayFormat := FScriptor.DateTimeFormat;
      END;
    end;
    FFieldNames.Objects[i] := anyField;
  end;

  ss := FScriptor.GetScriptState;
  FLoopBackLine := ss.linenum;
  dct := 1;

  {interpret succeeding lines until find ENDFOR;
  problem is succeeding lines may contain
  text that contains 'ENDFOR'}
  
  with ss do
  Repeat
    if (lineNum >= script.count) then
      break;
    xtoken := script[lineNum];
    r := StrToPChar(xtoken);
    r := StrStartToken(r);

    if (r^ = '@') then
    begin
      Inc(r);
      if StrNextTokenType(r,token, ptype, tokentype) then
      begin
        if COMPARETEXT(token,kForRows_F)=0 then
          Inc(dct);
        if COMPARETEXT(token,kEndForRows_F)=0 then
          Dec(dct);
      end;
    end;
    Inc(lineNum);
  Until(dct = 0);

  if (dct <> 0) then
    raise ERnnScript.Create('END_FOR_ROWS not found');

  FScriptor.SetScriptState(ss);{ready for next statement after end_for}
  FFirstPass := TRUE;
  if NextRow(ss) then
    begin
    ss.scriptlabel := kForRows_F;
    ss.refptr := SELF;
    FScriptor.PushScriptState(ss);
    Result := TRUE;
    end
end;

procedure ForRows_F(data:Pointer);
{@FOR [* | Fieldname1, Fieldname2, Fieldname#] in ':ALIAS:View.db'
	/* execute statements with variables set to field values for each row */
	delete from stqm where Fieldname1=$Fieldname1 and Fieldname2=$Fieldname2;
DONE}
var
  rowstep : TForRowsObj;
begin
  rowstep := TForRowsObj.Create;
  if not rowstep.InitLoop(PJvScriptCmdInfo(data)) then
    rowstep.Free;
end;

procedure EndForRows_F(data:Pointer);
var
  scriptor : TJvRnnBDEScript;
  sspop : TJvRnnScriptState;
  rowstep : TForRowsObj;
begin
  scriptor := PJvScriptCmdInfo(data)^.scriptor as TJvRnnBDEScript;
  sspop :=  scriptor.GetScriptState;

  if sspop.scriptlabel = kForRows_F then
  begin
    rowstep :=  sspop.refptr as TForRowsObj;
    if not rowstep.NextRow(sspop) then
      scriptor.PopScriptState
    else
      scriptor.SetScriptState(sspop);
  end
  else
    raise ERnnScript.create('END_FOR_ROWS without FOR_ROWS');
end;

procedure AbortOnZero_F(data:Pointer);
var
	varname, msg : string;

	test :integer;
	ptype : TJvRnnParenType;
    ix : integer;
    cmdInfo : TJvScriptCmdInfo;
	p : PChar;
    varray : TJvCCStrStrArray;
begin
    cmdInfo := TJvScriptCmdInfo(data^);
    p := cmdInfo.cmdcstr;
    varray := cmdInfo.scriptor.Variables;

	if not StrNextToken(p,varname,ptype) then
		raise ERnnScript.Create('@AbortOnZero: missing variable name');

	if (varname[1]='$') then
		varname := Copy(varname,2,Length(varname)-1);

	ix := varray.IndexOf(varname);
 	if (ix < 0) then
		raise ERnnScript.Create('No variable called '+varname);

	test := StrToIntDef(varray.GetIndexValue(ix),0);

    if (test = 0) then begin
    	msg := '';
    	if ( p<> nil) then
            msg := StrPas(p);
		raise ERnnScript.Create(msg);
	end;

end;

{$IFDEF WIN32}
procedure ReadView_F(data:Pointer);
var
  tblName : string;
  aliasName : string;
  token : string;
  db : TDatabase;
  i : integer;
  ptype : TJvRnnParenType;
  toktype : TJvRnnTokenType;
  cmdInfo : TJvScriptCmdInfo;
  p  : PChar;
  savep : PChar;
  scriptor : TJvRnnBDEScript;
  args : TStrings;
  locate : TStrings;
  rtab : TBDEDataset;
  fld : TField;
  s : string;
  isTable : boolean;
  v : variant;
  procedure ThrowError(const estr : string);
  var
    es :string;
  begin
    es := Format('Error: %s %s',[cmdInfo.cmdname,estr]);
    raise ERnnScript.Create(es);
  end;

  function NextArg : boolean;
  begin
    savep := StrStartToken(p);
    Result := StrNextTokenType(p,token,ptype,toktype);
  end;
  procedure CheckAssign;
  begin
    if args = nil then
      ThrowError('No ASSIGN arguments');
  end;
begin
  cmdInfo := TJvScriptCmdInfo(data^);
  p := cmdInfo.cmdcstr;
  scriptor := cmdInfo.scriptor as TJvRnnBDEScript;
  rtab := nil;
  args := nil;
  locate := nil;
  isTable := (Strcomp(cmdInfo.cmdname,kReadTable_F)=0);
  TRY
    while not NextArg do
    begin
      if ((toktype = tok_symbol) and (token[1]='('))
        OR ((toktype = tok_name) and (COMPARETEXT(token,'ASSIGN')=0)) then
      begin
        args := scriptor.ParseFunctionArgs(p);
        if args = nil then
          ThrowError('no arguments varname=field');
      end else
      if ((toktype = tok_name) and (COMPARETEXT(token,'LOCATE')=0)) then
      begin
        locate := scriptor.ParseFunctionArgs(p);
        if args = nil then
          ThrowError('no arguments field=value');
      end else
      if ((toktype = tok_name) and (COMPARETEXT(token,'SELECT')=0)) then
      begin  {must be at end of command}
        if isTable then
          ThrowError('select.. uses READ_QRY');
        CheckAssign;
        rtab := JvRnnUtBDE.DBOpenQuery(scriptor.FAliasDB,StrPas(savep));
      end else
      if isTable then
      begin {get the table name}
        p := savep;
        StrNextToken(p,token,ptype);
        GetAliasTableName(token,aliasName,tblName);
        if Length(aliasName) > 0 then
          db := scriptor.GetAliasedDB(aliasName)
        else
          db := scriptor.FAliasDB;
        if not DBTableExists(db,tblName) then
          ThrowError('Table not found: '+token);
        rtab := DBOpenTable(db, tblName);
      end else
      begin
        ThrowError('Unknown option');
      end;
    end;  {while}
    CheckAssign;
    if locate = nil then
    begin
      rtab.first;
      if rtab.eof then
        ThrowError('Empty dataset');
    end else
    begin
      if (locate.Count > 2) then
      begin
        v := VarArrayCreate([0,locate.Count-2],varVariant);
        for i := 1 to locate.count-1 do
        begin
          v[i-1] := locate[i];
        end;
      end else
      if (locate.Count = 2) then
      begin
        v := locate[1]
      end else
        ThrowError('LOCATE not enough arguments');
      if not rtab.Locate(locate[0], v, [] ) then
        ThrowError('LOCATE failed');
    end;
    for i := 0 to args.count-1 do
    begin
      token := GetStrName(args[i]);
      fld := rtab.FieldByName(args.Values[token]);
      CASE fld.DataType of
        ftString : s := QuoteStr(fld.AsString);
        ftDate : s := JvRnnStrUt.GetSQLDateStr(fld.AsDateTime);
        ftDateTime :
          begin
          if FRAC(fld.AsDateTime) <> 0 then
          begin
            if Scriptor.IsOracleDB(Scriptor.FAliasDB) then
              s := JvRnnUtBDE.OracleTimeStampStr(fld.AsDateTime)
            else
              s := JvRnnUtBDE.StdTimeStampStr(fld.AsDateTime);
          end
          else
            s := JvRnnStrUt.GetSQLDateStr(fld.AsDateTime);
          end;
        ftTime : s := QuoteStr(fld.AsString);
      ELSE
        s := fld.AsString;
      END{CASE};
      scriptor.SetVariable(token,s);
    end;
  finally
    rtab.free;
    locate.free;
    args.free;
  end;
end;
{$ENDIF WIN32}
procedure  Display_F(data:Pointer);
var
  tblName : string;
  aliasName : string;
  token : string;

  db : TDatabase;

  ptype : TJvRnnParenType;
  cmdInfo : TJvScriptCmdInfo;
  p  : PChar;
  scriptor : TJvRnnBDEScript;
begin
  cmdInfo := TJvScriptCmdInfo(data^);
  p := cmdInfo.cmdcstr;
  scriptor := cmdInfo.scriptor as TJvRnnBDEScript;

  if not StrNextToken(p,token,ptype) then
    raise ERnnScript.Create('@DISPLAY: missing table name');

  GetAliasTableName(token,aliasName, tblName);

  if Length(aliasName) > 0 then
    db := scriptor.GetAliasedDB(aliasName)
  else
    db := scriptor.FAliasDB;

  if DBTableExists(db,tblName) then
  begin
    WITH scriptor do
    begin
      ViewTableName := tblname;

      if Assigned (FAfterViewCreate) then
        FAfterViewCreate(cmdInfo.scriptor,db,tblName);

    end;
  end;
end;


procedure DeleteView_F(data:Pointer);
var
  tblName : string;
  aliasName : string;
  token : string;
  db : TDatabase;

  ptype : TJvRnnParenType;
  cmdInfo : TJvScriptCmdInfo;
  p  : PChar;
  scriptor : TJvRnnBDEScript;
begin
  cmdInfo := TJvScriptCmdInfo(data^);
  p := cmdInfo.cmdcstr;
  scriptor := cmdInfo.scriptor as TJvRnnBDEScript;

  if not StrNextToken(p,token,ptype) then
    raise ERnnScript.Create('@DELETE_QRY: missing table name');
  GetAliasTableName(token, aliasName, tblName);

  if Length(aliasName) > 0 then
    db := scriptor.GetAliasedDB(aliasName)
  else
    db := scriptor.FAliasDB;


  if DBTableExists(db,tblName) then with scriptor do
  begin
    if Assigned(FBeforeViewDestroy) then
      FBeforeViewDestroy(cmdInfo.scriptor,db,tblName);
    DeleteDBTable(db,tblName);
  end;
end;


procedure CreateViewOp(const cmdInfo : TJvScriptCmdInfo;
						const tblname:string; sql : PChar; keyct : integer);
var
  aliasname : string;
  rtabname : string;

  rtab : TTable;

  destDb : TDatabase;
  varray : TJvCCStrStrArray;
  scriptor : TJvRnnBDEScript;
begin
  scriptor := cmdInfo.scriptor as TJvRnnBDEScript;

  GetAliasTableName(tblName, aliasname, rtabname);

  if Length(aliasname) > 0 then
    destDb := scriptor.GetAliasedDB(aliasname)
  else
    destDb := scriptor.FAliasDB;

  varray := nil;
  WITH scriptor do
  begin
    if Variables <> nil then
      varray := Variables;
  end;

  rtab := InstantiateAnswerStr(scriptor.FAliasDB,destDb,sql,rtabname,keyct);
  try
    if varray <> nil then
    begin
      rtab.open;
      varray.SetVariable('RecordCount',IntToStr(rtab.RecordCount));
    end;
  finally
    rtab.free;
  end;
end;

procedure Query_F(data:Pointer);
var
  cmdInfo : PJvScriptCmdInfo;
  sql : PChar;
  db : TDatabase;
  scriptor : TJvRnnBDEScript;
begin
  cmdInfo := PJvScriptCmdInfo(data);
  sql := cmdInfo^.cmdcstr;
  scriptor := cmdInfo^.scriptor as TJvRnnBDEScript;
  db := scriptor.FPrivDB;
  CreateViewOp(cmdInfo^, ':'+db.DatabaseName+':Answer.db', sql, 0);
end;

procedure BatchView_F(data:Pointer);
var
  cmdInfo : TJvScriptCmdInfo;
  p : PChar;
  disp : PChar;
  resname, tblname : string;
  msg : string;

  ptype : TJvRnnParenType;
  sqlIsStored : boolean;
  aliasname : string;

  destdb : TDatabase;
  srcdb : TDatabase;
  scriptor : TJvRnnBDEScript;
  bmode : TBatchMode;
  sql : string;
  qry : TQuery;
  rtab : TTable;
  rct : integer;
  matchfieldnames : boolean;
begin
	{@Insert_QRY  tablename  sql }
 {@UPDATE_QRY  tablename  sql }
  cmdInfo := TJvScriptCmdInfo(data^);
  p := cmdInfo.cmdcstr;
  scriptor := cmdInfo.scriptor as TJvRnnBDEScript;
  
  if not StrNextToken(p,resname,ptype) then
    raise ERnnBDEScript.Create('@INSERT_QRY: missing destination name');

  sqlIsStored := (ptype = paren_angle);

  if not sqlIsStored then
    tblname  := resname
  else {next token is table name}
    if not StrNextToken(p,tblname, ptype) then
      {assume result is same database and name as query}
      tblname := resname;

  {after the table name, but before sql, can have USE_FIELDNAMES token}
  disp := p; {save p in case it points to select}
  StrNextToken(p, msg, ptype);
  matchfieldnames := (msg = 'USE_FIELDNAMES');
  if (not matchfieldnames) then
  begin
    if COMPARETEXT('select',msg) <> 0 then
      raise ERnnScript.create('select.. SQL expected');
    p := disp;
  end;

  disp := nil;
  try
    if not sqlIsStored then
    begin
    {rest is sql}
      while (p^ <> #0) AND (p^ <= #32) do
        Inc(p);
      sql := StrPas(p);
    end else
    begin
      sql := scriptor.ScriptStore.GetScriptText(resname);
    end;
  finally
    if disp <> nil then
      StrDispose(disp);
  end;

  {fix to add USE_FIELDNAMES capability for batch moves}
  {see if sql text starts with 'select'}

  GetAliasTableName(tblname,aliasname, resname);
  if Length(aliasname) > 0 then
    destDb := scriptor.GetAliasedDB(aliasname)
  else
    destDb := scriptor.FAliasDB;
  srcdb := Scriptor.FAliasDB;
  msg := StrPas(cmdInfo.cmdname);

  if COMPARETEXT(msg,'INSERT_QRY')=0 then
  begin
    if not DBTableExists(destdb,resname) then
    begin
      bmode := batCopy;
      matchfieldnames := FALSE;
    end
    else
      bmode := batAppend;
  end else if COMPARETEXT(msg,'DELETE_QRY')=0 then
  begin
    bmode := batDelete;
  end
  else begin       {UPDATE_QRY}
    bmode := batAppendUpdate;
  end;

  rtab := nil;
  qry := TQuery.Create(nil);
  TRY
    qry.DatabaseName := srcdb.DatabaseName;
    qry.sql.clear;
    qry.sql.Add(sql);
    qry.Active := TRUE;
    if bmode = batCopy then {create a table with no primary key}
    begin
      rtab := JvRnnUtBDE.InstantiateQueryDB(qry,destdb,resname,0);
      rct := rtab.RecordCount;
    end
    else begin
      rtab := DBOpenTable(destdb, resname);
      if (not matchfieldnames) then
        rct := rtab.BatchMove(qry,bmode)
      else
      begin
        rct := ApplyDatasetToTable(qry,rtab,nil,bmode,FALSE);
      end;
    end;
    scriptor.SetVariable('RecordCount',IntToStr(rct));
  FINALLY
    qry.free;
    rtab.free;
  END;

 (* batchMove := BatchMoveMode(srcdb,destdb,'TEMPVIEW.DB',resname,bmode)
  try
    if (batchMove.ProblemCount > 0) OR (batchMove.KeyViolCount > 0) then
    begin
      msg := Format('Append to %s: %d moved, %d problems, %d key violations',
       [resname, batchMove.MovedCount, batchMove.ProblemCount, batchMove.KeyViolCount]);
      scriptor.LogError(msg);
    end;
  finally
    batchMove.Free;
  end; *)
end;



procedure BDEConnectAlias_F(data:Pointer);
var
  cmdInfo : TJvScriptCmdInfo;
  p : PChar;
  scriptor : TJvRnnBDEScript;
  ptype : TJvRnnParenType;
  aliasname : string;
  params : TStrings;
  db : TDatabase;

  procedure ThrowError(const instr : string);
  begin
    raise ERnnScript.create('@CONNECT_ALIAS: '+instr);
  end;
begin
  cmdInfo := TJvScriptCmdInfo(data^);
  p := cmdInfo.cmdcstr;
  scriptor := cmdInfo.scriptor as TJvRnnBDEScript;

  if not StrNextToken(p,aliasname,ptype) then
    ThrowError('missing alias name');

  params := scriptor.ParseFunctionArgs(p);
  try
    db := Session.FindDatabase(aliasname);
    if (db <> nil) and (db.Connected) then
      db.Connected := FALSE;
    scriptor.ConnectAlias(aliasname, params);
  finally
    params.free;
  end;
end;

{$IFDEF RL_D3}
procedure BDEModifyAlias_F(data:Pointer);
var
  cmdInfo : TJvScriptCmdInfo;
  p : PChar;
  ptype : TJvRnnParenType;
  aliasname : string;
  params : TStrings;
  db : TDatabase;
  scriptor : TJvRnnCustomScript;

  procedure ThrowError(const instr : string);
  begin
    raise ERnnScript.create('@MODIFY_ALIAS: '+instr);
  end;

begin
  cmdInfo := TJvScriptCmdInfo(data^);
  p := cmdInfo.cmdcstr;

  if not StrNextToken(p,aliasname,ptype) then
    ThrowError('missing alias name');

  scriptor := TJvRnnCustomScript(cmdInfo.scriptor);
  params := scriptor.ParseFunctionArgs(p);
  try
    db := Session.FindDatabase(aliasname);
    if (db <> nil) and (db.Connected) then
      db.Connected := FALSE;

    Session.ModifyAlias(aliasname, params);
  finally
    params.free;
  end;
end;
{$ENDIF}

{$IFDEF RL_D3}
procedure BDEADDAlias_F(data:Pointer);
var
  cmdInfo : TJvScriptCmdInfo;
  p : PChar;
  ptype : TJvRnnParenType;
  aliasname : string;
  drivername : string;
  params : TStrings;

  procedure ThrowError(const instr : string);
  begin
    raise ERnnScript.create('@CREATE_ALIAS: '+instr);
  end;

begin
  cmdInfo := TJvScriptCmdInfo(data^);
  p := cmdInfo.cmdcstr;

  if not StrNextToken(p,aliasname,ptype) then
    ThrowError('missing alias name');

  if not StrNextToken(p,drivername,ptype) then
    ThrowError('missing driver name');

  params := cmdInfo.scriptor.ParseFunctionArgs(p);
  try
    Session.AddAlias(aliasname,drivername, params);
  finally
    params.free;
  end;
end;
{$ENDIF}

procedure CreateView_F( data:Pointer);
var
  token : string;
  tblname : string;
  keyct : integer;

  ptype : TJvRnnParenType;
  cmdInfo : TJvScriptCmdInfo;
  p : PChar;
  savep : PChar;
  tok_type : TJvRnnTokenType;

  function NextAnyToken : boolean;
  begin
    Result := StrNextToken(p,token,ptype);
  end;

  function NextTokenType : boolean;
  begin
    savep := p;
    Result := StrNextTokenType(p,token,ptype,tok_type);
  end;

  procedure ThrowErr;
  begin
    raise ERnnScript.Create(
      '@CREATE_QRY: [:alias:]<tablename> [<keyct>] [as] select..');
  end;

  function IsSelect : boolean;
  begin
    result := (COMPARETEXT('select',token)=0);
  end;

  function IsAS : boolean;
  begin
    result := (COMPARETEXT('as',token)=0);
  end;

  procedure AsSelect;
  begin
    if IsAS then
      if not NextTokenType then
        ThrowErr;
    if not IsSelect then
      ThrowErr;
    p := savep;
    while (p^<>#0) and (p^ <= ' ') do
      Inc(p);
  end;

begin
  {@CREATE_QRY  tablename <keyct>  }
  cmdInfo := TJvScriptCmdInfo(data^);
  p := cmdInfo.cmdcstr;
  {scriptor := cmdInfo.scriptor as TJvRnnBDEScript;}

  if not NextAnyToken then
    ThrowErr;
  tblname := token;
  if not NextTokenType then
    ThrowErr;

  if tok_type = tok_integer then
  begin
    keyct := StrToInt(token);
    if not NextTokenType then
      ThrowErr;
    AsSelect;
  end
  else begin
    keyct := 0;
    AsSelect;
  end;
  CreateViewOp(cmdInfo, tblname, p,  keyct);
end;

procedure TJvRnnBDEScript.ExecuteStd(cmd : TStrings);
begin
  FQuery.sql.Assign(cmd);
  if cmd.count = 0 then
    exit;
  try
    FQuery.Prepare;
    FQuery.ExecSQL;
    FQuery.close;
  except
    On ex1:EDatabaseError do
    begin
      LogMsg('Error Running ' + cmd[0] + ' : '+ex1.Message);
      raise;
    end;
  end;
end;

procedure TJvRnnBDEScript.SetAliasName(const aliasname:string);
begin
  FAliasDB := GetAliasedDB(aliasName); (*check DB is valid*)
  FAlias := aliasname;
  FQuery.SQL.Clear;
  FQuery.DatabaseName := FAlias;
end;

function TJvRnnBDEScript.GetAliasedDB(const inAlias:string):TDatabase;
begin
  Result := GetDatabaseFromAlias(inAlias);
  if Result.Owner = nil then
    InsertComponent(Result);
end;

function BDETableExists_F(sct:TJvRnnCustomScript; args : TStrings): string;
var
  db : TDatabase;
  scriptor : TJvRnnBDEScript;
  aliasname : string;
  tabname : string;
begin
  scriptor :=  sct as TJvRnnBDEScript;
  if (args.count <= 0) then
  begin
    Result := 'N';
    exit;
  end;
  GetAliasTableName(args[0],aliasname, tabname);
  if Length(aliasname) > 0 then
    db := scriptor.GetAliasedDB(aliasname)
  else
    db := scriptor.FAliasDB;
  if DBTableExists(db, tabname) then
    Result := 'Y'
  else
    Result := 'N';
end;

{$IFDEF RL_D3}
function BDEAliasExists_F(sct:TJvRnnCustomScript; args : TStrings): string;
var
  exists : boolean;
begin
  if (args.Count > 0) then
  begin
    exists := Session.IsAlias(args[0]);
  end
  else
    exists := FALSE;
  
  if (exists) then
    Result := 'Y'
  else
    Result := 'N';
end;
{$ENDIF}

end.
 
