{-----------------------------------------------------------------------------
The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in compliance
with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/MPL-1.1.html

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either expressed or implied. See the License for
the specific language governing rights and limitations under the License.

The Original Code is: JvRnnScIB.PAS, released on 2000-12-01.

The Initial Developer of the Original Code is Michael Rynn.
Portions created by Michael Rynn are Copyright (C) 2000  Michael Rynn.
All Rights Reserved.

Contributor(s): ______________________________________.

Last Modified: 2000-12-01

You may retrieve the latest version of this file at the Project JEDI home page,
located at http://www.delphi-jedi.org

-----------------------------------------------------------------------------}
{$A+,B-,C+,D+,E-,F-,G+,H+,I+,J+,K-,L+,M-,N+,O+,P+,Q-,R-,S-,T-,U-,V+,W-,X+,Y+,Z1}

{$I JCL.INC}
unit JvRnnScIB;


interface
Uses Classes,SysUtils,JvCCArray,JvRnnScrSt,JvRnnScCst,JvRnnVarCp,
  IBDatabase,IBSQL,IB;

type


  TJvRnnIBScript = class;

  ERnnIBScript = class(ERnnScript);

  TJvRnnIBScript = class(TJvRnnCustomScript)
  private
    FDatabase : TIBDatabase;
    FQuery : TIBSQL;

    FCommitDDL : boolean;
    FIsDDL : boolean;
    FWasCommit : boolean;
    FScriptTransaction : TIBTransaction;

    procedure SetDatabase(inDB : TIBDatabase);
    procedure CreateScriptDatabase(const dbpath:string; dbparams:TStrings);
    procedure OpenScriptDatabase(const dbpath:string; dbparams:TStrings);
    procedure DisconnectScriptDatabase;
  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;

    procedure CreateAndOpenDatabase(ps : PChar);
    procedure OpenDatabase(ps : PChar);
    procedure ExecuteStd( cmd : TStrings);  override;
    function  ParseSQL(cmd : TStrings): boolean;
    procedure CheckDatabaseComponent;

  public
    constructor Create(AOwner:TComponent);override;
    destructor Destroy;override;

    property Transaction : TIBTransaction read FScriptTransaction;

  published
    property CmdExtract;
    property CmdProcess;
    property CmdExecute;
    property OnExecuteStd;

    property OnLogMsg;
    property BeforeCmdExecute;
    property OnLogError;
    property OnLogStrings;

    property ParamChar;
    property PrefixChar;
    property Terminator;
    property StdTerminator;
    property ScriptStore;
    property ValueList;
    property Database : TIBDatabase read FDatabase write SetDatabase;
  end;

procedure Register;

implementation
Uses Forms,JvRnnStrUt,Dialogs,JvRnnStrLs,JvRnnPars;

constructor TJvRnnIBScript.Create(AOwner:TComponent);
begin
  inherited create(AOwner);

  FQuery := TIBSQL.Create(SELF);
  FScriptTransaction := TIBTransaction.Create(SELF);
  FScriptTransaction.Tag := 501;
  FQuery.Transaction := FScriptTransaction;

  InstallCustomScriptOps(SELF);
end;

destructor TJvRnnIBScript.Destroy;
begin
  FScriptTransaction.Active := FALSE;
  FUserOps.Free;
  inherited Destroy;
end;

procedure TJvRnnIBScript.CheckDatabaseComponent;
begin
  if FDatabase = nil then
    raise ERnnIBScript.create('No database assigned');
end;

procedure TJvRnnIBScript.Notification(AComponent: TComponent;
			Operation: TOperation);
begin
  if (Operation = opRemove) then
  begin
    if Database = AComponent then
      Database := nil
  end;
  inherited Notification(AComponent, Operation);
end;

procedure TJvRnnIBScript.SetDatabase(inDB : TIBDatabase);
var
  ix : integer;
begin
  if inDB = FDatabase then
    exit;
  if FDatabase <> nil then
  begin
    ix := FScriptTransaction.FindDatabase(FDatabase);
    if (ix >= 0) then
      FScriptTransaction.RemoveDatabase(ix);
  end;
  FDatabase := inDB;
  if FDatabase <> nil then
  begin { add 2-way association}
    FScriptTransaction.AddDatabase(FDatabase);
    FDatabase.AddTransaction(FScriptTransaction);
    if FDatabase.Connected then
      FScriptTransaction.Active := TRUE;
  end;
end;

procedure  TJvRnnIBScript.DisconnectScriptDatabase;
begin
  CheckDatabaseComponent;
  if FDatabase.Connected then
    FDatabase.Connected := FALSE;
end;

procedure  TJvRnnIBScript.OpenScriptDatabase(const dbpath:string; dbparams:TStrings);
begin
  DisconnectScriptDatabase;
  FDatabase.DatabaseName := dbpath;
  FDatabase.Params.Assign(dbparams);
  FDatabase.Connected := TRUE;
end;

procedure  TJvRnnIBScript.CreateScriptDatabase(const dbpath:string; dbparams:TStrings);
begin
  DisconnectScriptDatabase;
  FDatabase.DatabaseName := dbpath;
  FDatabase.Params.Assign(dbparams);
  FDatabase.CreateDatabase;
end;

function  TJvRnnIBScript.ParseSQL(cmd : TStrings): boolean;
var
  cs : PChar;
  p : PChar;
  token : string;
  ptype : TJvRnnParenType;
  toktype : TJvRnnTokenType;
  temp : boolean;

  function TokenEquals(const invalue : string) : boolean;
  begin
    Result := (COMPARETEXT(token, invalue) = 0);
  end;

  function GotNextToken : boolean;
  begin
    Result := StrNextTokenType( p, token, ptype, toktype) and (toktype=tok_name);
  end;

begin
	 {check for SET TERM, CONNECT and CREATE DATABASE parameters}
  cs := StrListToCStr(cmd,' ');
  try
    Result := FALSE;
    FIsDDL := FALSE;
    FWasCommit := FALSE;

    p := cs;
    if GotNextToken and (ptype=paren_none) then
    begin
      if TokenEquals('SET') then
      begin
        if GotNextToken then
        begin
          if TokenEquals('TERM') then
          begin
            {expect ASCII printable character for new statement terminator}
            if StrNextTokenType( p, token, ptype, toktype)
              and (toktype=tok_bad) then
            begin
            {not a name or math op or relation op, must be a character?}
              StdTerminator := token[1];
              Result := TRUE;
            end
          end
          else if TokenEquals('AUTO')
            OR TokenEquals('AUTODDL') then
          begin
            {expect ON or OFF}
            temp := FALSE;
            if GotNextToken then
            begin
              if TokenEquals('ON') then
              begin
                temp := TRUE;
                FCommitDDL := TRUE;
              end else if TokenEquals('OFF') then
              begin
                temp := TRUE;
                FCommitDDL := FALSE;
              end;
            end;
           if not temp then
             raise ERnnScript.create('ON or OFF expected');
           Result := TRUE;
        end;
      end;
    end else if TokenEquals('CONNECT') then
    begin
      OpenDatabase(p);
      Result := TRUE;
    end else if TokenEquals('CREATE') then
    begin
      if GotNextToken and TokenEquals('DATABASE') then
      begin
        CreateAndOpenDatabase(p);
        Result := TRUE;
      end else
        FIsDDL := TRUE;
    end else if TokenEquals('ALTER') OR TokenEquals('DROP') OR
      TokenEquals('GRANT') OR TokenEquals('REVOKE') then
      FIsDDL := TRUE
    else if TokenEquals('COMMIT') then
    begin
      FWasCommit := TRUE;
    end;
  end;
  finally
    StrDispose(cs);
  end;
end;

procedure TJvRnnIBScript.OpenDatabase(ps : PChar);
var
  dbparams : TJvRnnStrList;
  token : string;
  p : PChar;
  ptype : TJvRnnParenType ;
  toktype : TJvRnnTokenType;
  dbname : string;

  function NextToken : boolean;
  begin
    Result := StrNextTokenType( p, token, ptype, toktype) and (toktype<>tok_bad);
  end;

  procedure ValueExpected(const inOption : string);
  begin
    raise ERnnIBScript.create('Value expected for '+inOption);
  end;

  procedure CheckOption(const inOption:string);
    procedure  sub_getvalue;
    begin
      if not NextToken then
        ValueExpected(inOption);
      if toktype = tok_equal then
        if not NextToken then
          ValueExpected(inOption);
    end;

  begin
    if (COMPARETEXT(inOption,'USER')=0) then
    begin
      sub_getvalue;
      dbparams.Values['user_name'] := token;
    end else if (COMPARETEXT(inOption,'PASSWORD')=0) then
    begin
      sub_getvalue;
      dbparams.Values['password'] := token;
    end else
      raise ERnnIBScript.create('Unknown option '+inOption);
  end;

begin
  p := ps;
  if NextToken then
  begin
    dbname := token;
    {translate USER and PASSWORD options}
    dbparams := TJvRnnStrList.Create;
    try
      while NextToken do
        CheckOption(token);
      OpenScriptDatabase(dbname,dbparams);
    finally
      dbparams.free;
    end;
  end;
end;

procedure TJvRnnIBScript.CreateAndOpenDatabase(ps : PChar);
var
  p : PChar;
  token : string;
  ptype : TJvRnnParenType ;
  toktype : TJvRnnTokenType;
  dbname : string;
  dbparams : TJvRnnStrList;

  function NextToken : boolean;
  begin
    Result := (StrNextTokenType( p, token, ptype, toktype) and ((toktype=tok_name)or(toktype=tok_quotetext)));
  end;

begin
  {CREATE DATABASE "?DB_PATH" PAGE_SIZE 4096 USER "SYSDBA" PASSWORD "masterkey" }
  p := ps;
  if NextToken  then
  begin
    dbname := token;
    dbparams := TJvRnnStrList.create;
    try
      dbparams.Add(StrPas(p));
      CreateScriptDatabase(dbname,dbparams);
    finally
      dbparams.free;
    end;
  end;
end;


procedure TJvRnnIBScript.ExecuteStd(cmd : TStrings);
begin
  if ParseSQL(cmd) then
    exit;

  if FQuery.Open then
    FQuery.CheckClosed;
  if cmd.count=0 then exit;
  
  FQuery.Database := FDatabase;
  FQuery.SQL.Assign(cmd);
  try
    FScriptTransaction.Active := TRUE;  {if not already active}
    FQuery.Prepare;
    FQuery.ExecQuery;
    FQuery.Close;
    if (FWasCommit) OR ((FIsDDL) and (FCommitDDL)) then
    begin
      if (FScriptTransaction.InTransaction) then
        FScriptTransaction.Commit;     {supports auto-commit DDL?}
    end;
  except
    On ex1:EIBError do
    begin
      LogMsg('Error Running ' + cmd[0] + ' : '+ex1.Message);
      raise;
    end;
  end;
end;



procedure Register;
begin
  RegisterComponents('RnnLib', [TJvRnnIBScript]);
end;  

begin
	//RegisterClasses([TJvRnnIBScript]);
end.




