{-----------------------------------------------------------------------------
The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in compliance
with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/MPL-1.1.html

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either expressed or implied. See the License for
the specific language governing rights and limitations under the License.

The Original Code is: RnnVarEd.PAS, released on 2000-12-01.

The Initial Developer of the Original Code is Michael Rynn.
Portions created by Michael Rynn are Copyright (C) 2000  Michael Rynn.
All Rights Reserved.

Contributor(s): ______________________________________.

Last Modified: 2000-12-01

You may retrieve the latest version of this file at the Project JEDI home page,
located at http://www.delphi-jedi.org

-----------------------------------------------------------------------------}
{$A+,B-,C+,D+,E-,F-,G+,H+,I+,J+,K-,L+,M-,N+,O+,P+,Q-,R-,S-,T-,U-,V+,W-,X+,Y+,Z1}

{$I JCL.INC}
unit JvRnnVarEd;

interface

uses
  WinProcs, WinTypes, Messages, SysUtils, Classes, Graphics,Controls,
  Forms, Dialogs, ExtCtrls, JvCCArray, Buttons, StdCtrls, Grids;

type

  TfrmRnnStrStrArray = class(TForm)
    Panel1: TPanel;
    Panel2: TPanel;
    btnAdd: TSpeedButton;
    btnDelete: TSpeedButton;
    btnOK: TButton;
    btnCancel: TButton;
    sgrid: TDrawGrid;

     procedure FormCreate(Sender: TObject);
    procedure btnAddClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure btnDeleteClick(Sender: TObject);
    procedure btnOKClick(Sender: TObject);
    procedure sgridDrawCell(Sender: TObject; Col, Row: Longint;
      Rect: TRect; State: TGridDrawState);
    procedure sgridGetEditText(Sender: TObject; ACol, ARow: Longint;
      var Value: OpenString);
    procedure sgridSetEditText(Sender: TObject; ACol, ARow: Longint;
      const Value: String);
  private
   { Private declarations }
    FRyArray : TJvCCStrStrArray;
    FTextData : string;
    FEditCol : LongInt;
     FEditRow : LongInt;


  public
   { Public declarations }
     procedure SetEditData(const s : string; Col,Row : integer);
     procedure WndProc(var amsg: TMessage); override;
    procedure StopEditing(saveit:boolean);
    procedure SetVariables(svar:TJvCCStrStrArray);
    procedure GetVariables(svar:TJvCCStrStrArray);
    procedure NewItem;
    procedure RefreshVarDisplay;
  end;

const
  kSetActiveRow = 'RnnSSPropSetActiveRow';
var
  frmRnnStrStrArray: TfrmRnnStrStrArray;
  gVarPropSetRow : WORD;

function EditStrStrArray(slist : TJvCCStrStrArray):boolean;

implementation

const
  kRowNumCol = 0;
  kNameCol = 1;
  kValueCol = 2;

{$R *.DFM}


function EditStrStrArray(slist : TJvCCStrStrArray):boolean;
var
  eform : TfrmRnnStrStrArray;
begin
  if slist is TJvCCStrStrArray then
  begin
  eform := TfrmRnnStrStrArray.Create(nil);
     try
      eform.SetVariables(slist);
      result := (eform.ShowModal=mrOK);
        if result then
      begin
           eform.GetVariables(slist);
        end;
    finally
        eform.free;
     end;
   end else
      result := false;
end;

procedure TfrmRnnStrStrArray.StopEditing(saveit:boolean);
var
  ix : integer;
begin
  if (FEditRow <= 0) OR (FEditCol <= 0) then
      exit;
  ix := FEditRow - 1;

  if saveit then
  CASE FEditCol of
    kNameCol :
      begin
        FRyArray.SetIndexName(ix,FTextData);
        FRyArray.QuickSort;
        sgrid.Invalidate;
        ix := FRyArray.IndexOf(FTextData);
        if ix >= 0 then
          PostMessage(SELF.Handle,gVarPropSetRow,0,ix+1);
      end;
    kValueCol :
      begin
        FRyArray.SetIndexValue(ix,FTextData);
      end;
  END;{CASE}
  FEditRow := 0;
  FEditCol := 0;
end;

procedure TfrmRnnStrStrArray.NewItem;
var
  s : string;
  varray : TJvCCStrStrArray;
begin
  StopEditing(TRUE);
  s := 'Unnamed';
  if Length(s) > 0 then
  begin
    varray := FRyArray;
    if (varray.IndexOf(s) >= 0) then begin
      MessageDlg('Variable '+s+ ' exists already',
      mtError,[mbOK],0);
      exit;
    end;
  varray.SetVariable(s,'');
  RefreshVarDisplay;
  end;
end;

procedure TfrmRnnStrStrArray.SetVariables(svar:TJvCCStrStrArray);
begin
  FRyArray.Assign(svar);
  RefreshVarDisplay;
end;

procedure TfrmRnnStrStrArray.GetVariables(svar:TJvCCStrStrArray);
begin
  svar.Assign(FRyArray);
end;

procedure TfrmRnnStrStrArray.FormCreate(Sender: TObject);
begin
  FRyArray := TJvCCStrStrArray.Create(0);
  RefreshVarDisplay;
end;

procedure TfrmRnnStrStrArray.btnAddClick(Sender: TObject);
begin
  NewItem;
end;


procedure TfrmRnnStrStrArray.RefreshVarDisplay;
var
  rowcount : integer;
begin
  rowcount := FRyArray.Count;
  if rowcount = 0 then
    rowcount := 1;
  sgrid.RowCount := rowcount+1;
  sgrid.Refresh;
end;

procedure TfrmRnnStrStrArray.SetEditData(const s : string; Col,Row : integer);
begin
  if (Row <= 0) OR (Col <= 0) then
    exit;
  if (FEditRow > 0) AND ((Row <> FEditRow)OR(Col<>FEditCol)) then
    StopEditing(TRUE);
  FEditRow := Row;
  FEditCol := Col;
  FTextData := s;
end;




procedure TfrmRnnStrStrArray.WndProc(var amsg: TMessage);
begin
  if amsg.Msg = gVarPropSetRow then
  begin
    sgrid.Row := amsg.LParam;
  end else
  begin
    inherited WndProc(amsg);
  end;
end;


procedure TfrmRnnStrStrArray.FormDestroy(Sender: TObject);
begin
  FRyArray.Free;
end;

procedure TfrmRnnStrStrArray.btnDeleteClick(Sender: TObject);
var
  row : LongInt;
  col : LongInt;
begin
  StopEditing(FALSE);
  row := sgrid.Row;
   col := sgrid.Col;

   if (row<=0)or(col<=0) then
    exit;
   FRyArray.DeleteItemsAt(row-1,1);
   RefreshVarDisplay;
end;

procedure TfrmRnnStrStrArray.btnOKClick(Sender: TObject);
begin
  StopEditing(TRUE);
end;





procedure TfrmRnnStrStrArray.sgridDrawCell(Sender: TObject; Col,
  Row: Longint; Rect: TRect; State: TGridDrawState);
var
  cv : TCanvas;
  outs : string;
  ix : integer;
const
  kVOffset = 3;
begin
  cv := sgrid.Canvas;
  if (Row>0) AND (Col=FEditCol) AND (Row=FEditRow) then
  begin
    StopEditing(TRUE);
  end;

  CASE col of
    0 :
      begin
        if (Row > 0) then
        begin
          outs := IntToStr(Row);
          cv.TextOut(Rect.Left+2,Rect.Top+kVOffset,outs);
        end;
      end;
    kNameCol :
      begin
        if Row=0 then
        begin
          cv.TextOut(Rect.Left+2,Rect.Top+kVOffset,'Name');
        end else
        begin
          ix := Row - 1;

          if (ix < 0) OR (ix >= FRyArray.Count) then
            exit;
          outs := FRyArray.GetIndexName(ix);
          cv.TextOut(Rect.Left+2,Rect.Top+kVOffset,outs);
        end;
      end;

    kValueCol :
      begin
        if Row=0 then begin
          cv.TextOut(Rect.Left+2,Rect.Top+kVOffset,'Value');
        end else
        begin
          ix := Row - 1;

          if (ix < 0) OR (ix >= FRyArray.Count) then
            exit;
          outs := FRyArray.GetIndexValue(ix);
          cv.TextOut(Rect.Left+2,Rect.Top+kVOffset,outs);
        end;
      end;
    END; {CASE}
end;

procedure TfrmRnnStrStrArray.sgridGetEditText(Sender: TObject; ACol,
  ARow: Longint; var Value: OpenString);
var
  ix : integer;

begin
  CASE ACol of
    0 :
      begin
      end;
    1 :
      begin
        ix := ARow - 1;
        if (ix >= FRyArray.Count) then
          exit;
        Value := FRyArray.GetIndexName(ix);
      end;
    2 :
      begin
      ix := ARow - 1;
      if (ix >= FRyArray.Count) then
        exit;
      Value := FRyArray.GetIndexValue(ix);
      end;
  END; {CASE}
end;

procedure TfrmRnnStrStrArray.sgridSetEditText(Sender: TObject; ACol,
  ARow: Longint; const Value: String);
var
  ix : integer;
begin
  if (ARow<=0) OR (ACol <= 0) then
    exit;
  ix := ARow-1;
  if (ix >= FRyArray.Count) then
    exit;
  SetEditData(Value,ACol,ARow);
end;

Initialization
   gVarPropSetRow :=  RegisterWindowMessage(kSetActiveRow);

end.
