unit JvRnnXProc;

interface
{$I JCL.INC}
uses
  Windows, Messages, SysUtils, Classes, Dialogs;

type
  ERnnXProcess = class(Exception);

  TJvRnnXProcess = class;

  TJvRnnXProcessWait = class(TThread)
  private
    FXProcess : TJvRnnXProcess;
    FRunning : boolean;
  public
    constructor Create(xprocess:TJvRnnXProcess);
    procedure   CheckRunning;
    procedure   ProcessEnded;
    procedure   Execute; override;
  end;

  TJvRnnXWaitTime = procedure(Sender:TJvRnnXProcess; var sleep_ms:integer) of object;

  TJvRnnXProcess = class(TComponent)
  private
    FStartupTimeOut:DWORD;
    FWaitForInput: integer;

    FStopOnDestroy:boolean;
    FStartOnCreate:boolean;
    FOnStarted:TNotifyEvent;
    FMainWindowClassName:string;
    FEXEName:string;
    FStartupDirectory:string;
    FCommandLine:string;
    FProcessInfo : TProcessInformation;
    FLastExitCode : integer;
    FFirstExitCode : integer;
    FWaitTillExit : boolean;
    FUseThread : boolean;
    FCanCheckProcess : boolean;
    FWaitTimedOut : boolean;
    FForget : boolean;

    FProcessEnded : TNotifyEvent;
    FWaitTime : TJvRnnXWaitTime;
    FThread : TJvRnnXProcessWait;

    procedure SetStartupDirectory(Value:string);
    procedure SetEXEName(Value:string);
    procedure SetMainWindowClassName(Value:string);
    procedure SetForget(doForget : boolean);
    procedure SetUseThread(trackThread:boolean);
  protected
    procedure DoOnStarted;

    function GetHandle:HWnd; virtual;
    function GetRunning:boolean; virtual;
    procedure SetRunning(Value:boolean); virtual;
    function StartProgram(State:integer):boolean;
    procedure NotifyEnd;
    procedure SetThread(waitThread:TJvRnnXProcessWait);
  public
    constructor Create(AOwner:TComponent); override;
    Destructor Destroy; override;

    procedure GetProcessExitCode;
    procedure PostQuitToProcessWindow;
    procedure ForgetProcess;

    property Handle:HWnd read GetHandle;
    
  published
    property LastExitCode : integer read FLastExitCode;
    property Running:boolean read GetRunning write SetRunning;
    property EXEName:string read FEXEName write SetEXEName;
    property MainWindowClassName:string read FMainWindowClassName write SetMainWindowClassName;
    property StartupDirectory:string read FStartupDirectory write SetStartupDirectory;
    property OnStarted:TNotifyEvent read FOnStarted write FOnStarted;

    property CommandLine:string read FCommandLine write FCommandLine;
    property Thread :  TJvRnnXProcessWait read FThread write SetThread;
    property StopOnDestroy:boolean read FStopOnDestroy write FStopOnDestroy;
    property StartOnCreate:boolean read FStartOnCreate write FStartOnCreate;
    property WaitForInputTime : integer read FWaitForInput write FWaitForInput default 3000;
    property StartupTimeOut: DWORD read FStartupTimeOut write FStartupTimeOut;
    property WaitTimedOut : boolean read FWaitTimedOut;
    property WaitTillExit : boolean read FWaitTillExit write FWaitTillExit default FALSE;
    property UseThread : boolean read FUseThread write SetUseThread default FALSE;
    property Forget : boolean read FForget write SetForget default FALSE;
    property OnProcessEnd : TNotifyEvent read FProcessEnded write FProcessEnded;
    property OnWaitTime :  TJvRnnXWaitTime read FWaitTime write FWaitTime;
  end;

implementation

uses
{$IFDEF DEBUG_LOG}
  UDebug,
{$ENDIF}
  Forms;

constructor TJvRnnXProcess.Create(AOwner:TComponent);
begin
  inherited Create(AOwner);
  FStartupTimeOut:=1000;
  FWaitForInput := 3000;

  if FStartOnCreate and (not (csDesigning in ComponentState)) then
    Running:=True;
end;

destructor TJvRnnXProcess.Destroy;
begin
  if FThread <> nil then
    FThread.Terminate;
  if FStopOnDestroy then
    Running:=False;
  inherited Destroy;
end;

function GetFilePath(const FileName:string):string;
var
   pchFullPath, x:PChar;
begin
  GetMem(pchFullPath, MAX_PATH+1);
  TRY
  if SearchPath(nil, PChar(FileName), nil, MAX_PATH, pchFullPath, x)>0 then
    Result:=StrPas(pchFullPath)
  else
    Result:=FileName;
  FINALLY
    FreeMem(pchFullPath, MAX_PATH+1);
  END;
end;

procedure TJvRnnXProcess.SetForget(doForget:boolean);
begin
   FForget := doForget;
   if FForget then
     ForgetProcess;
end;

procedure TJvRnnXProcess.SetUseThread(trackThread:boolean);
begin
   FUseThread := trackThread;
   if (not FUseThread) and (FThread <> nil) then
     FThread.Terminate;
end;

procedure  TJvRnnXProcessWait.CheckRunning;
begin
  FRunning := FXProcess.Running;
end;


constructor TJvRnnXProcessWait.Create(xprocess:TJvRnnXProcess);
begin
  FXProcess := xprocess;
  FreeOnTerminate := TRUE;
  inherited Create(FALSE);
end;

procedure TJvRnnXProcessWait.ProcessEnded;
begin
  FXProcess.NotifyEnd;
end;

procedure TJvRnnXProcessWait.Execute;
begin
  repeat
    Sleep(2000);
    Synchronize(CheckRunning);
  until(Terminated or (not FRunning));
  if (not FRunning) and (not Terminated) then
    Synchronize(ProcessEnded);
end;

procedure TJvRnnXProcess.NotifyEnd;
begin
  if (FThread <> nil)  then
    FThread := nil;
  if Assigned(FProcessEnded) then
    FProcessEnded(SELF);
end;

function TJvRnnXProcess.StartProgram(State:integer):boolean;
var
  StartUpInfo:TStartUpInfo;
  appName : PChar;
  startDir : PChar;
  cmdLine : PChar;
  wresult : dword;
  FileName : string;
  CurrentDirectory : string;
begin
  CurrentDirectory := FStartupDirectory;
  if Length(FEXEName) > 0 then
  begin
    if Length(CurrentDirectory)=0 then
    begin
      FileName:=GetFilePath(FEXEName);
      CurrentDirectory:=ExtractFilePath(FileName);
    end;
  end;
  FillChar(StartUpInfo,SizeOf(TStartupInfo),0);
  FillChar(FProcessInfo,SizeOf(TProcessInformation),0);
  with StartupInfo do
  begin
    cb:=SizeOf(StartUpInfo);
    dwFlags:=STARTF_USESHOWWINDOW or STARTF_FORCEONFEEDBACK;
    wShowWindow:=State;
  end;
  if Length(FileName)=0 then
    appName := nil
  else
    appName := PChar(FileName);
  if Length(CurrentDirectory)=0 then
    startDir := nil
  else
    startDir := PChar(CurrentDirectory);
  if Length(CommandLine)=0 then
    cmdLine := nil
  else
    cmdLine := PChar(CommandLine);
  Result:=CreateProcess(appName, cmdLine, nil, nil, False, NORMAL_PRIORITY_CLASS,
    nil, startDir, StartupInfo, FProcessInfo);
  if Result then
  begin
    with FProcessInfo do
    begin
      wresult := WaitForInputIdle(hProcess,FWaitForInput);
      FWaitTimedOut := (wresult <> 0)
    end;
  end;
end;

procedure TJvRnnXProcess.SetThread(waitThread:TJvRnnXProcessWait);
begin
  if FThread <> nil then
  begin
    FThread.FreeOnTerminate := TRUE;
    FThread.Terminate;
  end;
  FThread := waitThread;
end;


procedure TJvRnnXProcess.GetProcessExitCode;
var
  pcode : DWORD;
begin
  if (not GetExitCodeProcess(FProcessInfo.hProcess,pcode)) then
  begin
    FCanCheckProcess := FALSE;
    ShowMessage('No exit code'+IntToStr(exitcode));
  end else
  begin
    if FFirstExitCode=0 then
      FFirstExitCode := pcode;
    FLastExitCode := pcode;
  end;
  if FLastExitCode <> STILL_ACTIVE then
    FCanCheckProcess := FALSE;
end;


procedure TJvRnnXProcess.DoOnStarted;
var
  waitms : integer;
begin
  if Assigned(FOnStarted) then
    FOnStarted(Self);
  if UseThread and (FThread = nil) then
  begin
    FThread := TJvRnnXProcessWait.Create(SELF);
  end;
  if (WaitTillExit) and (not UseThread) then
  begin
    waitms := 200;
    while Running do
    begin
      if Assigned(FWaitTime) then
        FWaitTime(SELF,waitms);
      Sleep(waitms);
    end;
  end;
end;

procedure TJvRnnXProcess.SetStartupDirectory(Value:string);
begin
  Running:=False;
  FStartupDirectory:=Value;
end;

procedure TJvRnnXProcess.SetEXEName(Value:string);
begin
  Running:=False;
  FEXEName:=Value;
end;

procedure TJvRnnXProcess.SetMainWindowClassName(Value:string);
begin
  Running:=False;
  FMainWindowClassName:=Value;
end;

function TJvRnnXProcess.GetHandle:HWnd;
begin
  Result:=FindWindow(PChar(FMainWindowClassName), PChar(#0));
end;

function TJvRnnXProcess.GetRunning:boolean;
begin
  if (FCanCheckProcess) and (FProcessInfo.hProcess <> 0) then
  begin
    GetProcessExitCode;
    Result := FCanCheckProcess;
  end
  else
    Result := FALSE;
end;

procedure  TJvRnnXProcess.PostQuitToProcessWindow;
var
  w : HWND;
begin
  if (Length(FMainWindowClassName) > 0) then
  begin
    w := GetHandle;
    if (w > 0) then
      PostMessage(w,WM_QUIT,0,0);
  end;
end;


procedure TJvRnnXProcess.SetRunning(Value:boolean);
const
  KIP_TIME=100; //so we don't hog thred time while we are waiting
var
{$IFDEF Delphi5}
  t:DWORD;
{$ELSE}
  t:longint;
{$ENDIF}
  w : HWND;
  eword : DWORD;
  msg : string;

begin
  Application.ProcessMessages;
  if Value then
  begin
    if (Running) and (WaitTillExit) then
      raise ERnnXProcess.create('Component is running a process already');

    FFirstExitCode := 0;
{$IFDEF DEBUG_LOG}
    msg := TimeToStr(Now);
    Writeln(gLog,'Start process ',FEXEName,FCommandLine, ' at ',msg);
{$ENDIF}

    if StartProgram(SW_SHOWNORMAL) then
    begin
    //make sure it is running
    {$IFDEF DEBUG_LOG}
      if FProcessInfo.hProcess=0 then
        Writeln(gLog,'No process handle');
    {$ENDIF}
      FCanCheckProcess := TRUE;
      Application.ProcessMessages;

      TRY
        sleep(KIP_TIME);  //have a kip
        t:=GetTickCount;
        while (not Running) and (t+FStartupTimeOut > GetTickCount) do
        begin
          sleep(KIP_TIME);  //have a kip
          Application.ProcessMessages;
        end;
		      //if it is still not running  then it must have croaked in startup
        if not FCanCheckProcess then
        begin
          msg := Format('Process %s %s started but has exited. '#13#10+
            'First exit code = %d, Last exit code = %d',
            [FEXEName, CommandLine, FFirstExitCode, FLastExitCode]);

{$IFDEF DEBUG_LOG}
          Writeln(gLog,msg);
{$ENDIF}
          Raise ERnnXProcess.Create(msg);
        end;
        DoOnStarted;
      FINALLY
        if FForget then
          ForgetProcess;
      END;
    end else
    begin
      eword := GetLastError();
      msg := Format('Process %s %s failed to start: Last error = %d',
        [FEXEName, CommandLine, eword]);
      raise ERnnXProcess.Create(msg);
    end;
  end else
  begin
    w := Handle;
    if (w > 0) then
      PostMessage(w, WM_QUIT, 0, 0);
  end;
end;

procedure  TJvRnnXProcess.ForgetProcess;
begin
  if FCanCheckProcess then
  begin
    FCanCheckProcess := FALSE;
    CloseHandle(FProcessInfo.hProcess);
    CloseHandle(FProcessInfo.hThread);
    FillChar(FProcessInfo,SizeOf(TProcessInformation),0);
    if (FThread <> nil) then
      FThread.Terminate;
  end;
end;

end.
