unit JvSplatFile;

interface
uses Classes, SysUtils, JvDirList, WinProcs,WinTypes;
type

  EJvSplatError = class(Exception);

  PJvSplatProgress = ^TJvSplatProgress;
  TJvSplatNotify = procedure (psp:PJvSplatProgress) of object;
  TJvSplatErrorLog = procedure (const s : string) of object;
  TJvSplatProgress = record
    filename : string[31];
    timeStart : TDateTime;
    fileStart : TDateTime;
    bytesDone : LongInt;
    bytesLeft : LongInt;
    retries : integer;
    fileNumber : LongInt;
    totalFiles : LongInt;
    totalBytes : LongInt;
    totalDone : LongInt;
    notify : TJvSplatNotify;
    errorLog : TJvSplatErrorLog;
  end;

  TJvSplatDirectory = class(TObject)
  private
    FSource : string;
    FDest : string;
    FRecurse : boolean;
    FConditions : integer;
    FFilter : string;
    FProg : PJvSplatProgress;
    FTotalFiles : LongInt;
    FTotalBytes : LongInt;
    FCurDest : string;
  protected
    constructor Create(const src, dest, filter : string;
      conditions : integer;
      recurse : boolean;
      sprog:PJvSplatProgress);
    procedure Execute;
    procedure AddCopyRec(const crec : TCopyRec);
    procedure CountCopyRec(const crec : TCopyRec);
  end;


const
	gMaxSplatRetries : integer = 3;
{$IFNDEF WIN32}
  MAX_PATH = 144;
type
  DWORD = LONGINT;

function ExtractFileDrive(const path : string) : string;
{$ENDIF}

{$IFDEF WIN32}
type

  TVolumeInfo = record
    rootPath : string;
    name : string;
    serial : DWORD;
    maxname : DWORD;
    sysflags : DWORD;
    sysname : string;
  end;



procedure GetVolumeInfo(const vspec : string; var vi : TVolumeInfo);
{$ENDIF}
procedure InitSplatProgress(var sprog : TJvSplatProgress; sn:TJvSplatNotify);
procedure DoSplatFile(const srcPath, destPath : string; sprog:PJvSplatProgress);
procedure DoSplatDirectory(const srcPath, destPath : string;
  sprog:PJvSplatProgress;conditions : integer;
  recurse : boolean);

function RuboutFile(const path : string) : boolean;
function GetTempNameInDirectory(const inName : string; maxnamelen : integer) : string;

function GetFileLength(const fname : string) : integer;
function GetAttrW32Str(dw:DWORD) : string;

implementation
Uses JvRnnStrUt,FileCtrl,Dialogs,JvCCRecrd,JvCCArray;


function GetFileLength(const fname : string) : integer;
var
  s : TSearchRec;
begin
  if (FindFirst(fname, faAnyFile, s)=0) then
  begin
    Result := s.Size;
    SysUtils.FindClose(s);
  end
  else
    Result := -1;
end;

{$IFDEF WIN32}
procedure GetVolumeInfo(const vspec : string; var vi : TVolumeInfo);
const
  vbuf = 256;
var
  nameb : array[0..vbuf-1] of CHAR;
  sysb : array[0..vbuf-1] of CHAR;
begin
  if GetVolumeInformation( StrToPChar(vspec),nameb,vbuf,
    @vi.serial, vi.maxname, vi.sysflags, sysb, vbuf) = TRUE then
  begin
    vi.rootPath := vspec;
    vi.name := string(nameb);
    vi.sysname := string(sysb);
  end;
end;
{$ENDIF}
constructor TJvSplatDirectory.Create(const src, dest, filter : string;
  conditions : integer;
  recurse : boolean;
  sprog:PJvSplatProgress);
begin
  FSource := src;
  FDest := dest;
  FFilter := filter;
  FConditions := conditions;
  FRecurse := recurse;
  FProg := sprog;
end;

procedure TJvSplatDirectory.Execute;
begin
  FTotalFiles := 0;
  FTotalBytes := 0;
  if FProg <> nil then
  begin
    CompareWalkDirectories(FSource,FDest,FFilter,FConditions,
      CountCopyRec,FRecurse);
    WITH FProg^ do
    begin
      totalBytes := FTotalBytes;
      totalFiles := FTotalFiles;
      fileNumber := 0;
      totalDone := 0;
      timeStart := Now;
    end;
  end;
  CompareWalkDirectories(FSource,FDest,FFilter,FConditions,
    AddCopyRec,FRecurse);
end;

procedure TJvSplatDirectory.AddCopyRec(const crec : TCopyRec);
var
  srcfile, destfile, path : string;
begin
  srcfile := FSource + kDirSeparator + GetCCStr(crec.subpath);
  destfile := FDest + kDirSeparator + GetCCStr(crec.subpath);
  path := ExtractFilePath(destFile);
  if (path <> FCurDest) then
  begin
    FCurDest := path;
    if not DirectoryExists(FCurDest) then
      ForceDirectories(FCurDest);
  end;
  if FProg <> nil then
    Inc(FProg^.filenumber);
  DoSplatFile(srcfile,destfile,FProg);
  if FProg <> nil then
    Inc(FProg^.totalDone,crec.v1.size);
end;

procedure TJvSplatDirectory.CountCopyRec(const crec : TCopyRec);
begin
  Inc(FTotalFiles);
  FTotalBytes := FTotalBytes + crec.v1.size;
end;

function GetAttrW32Str(dw:DWORD) : string;
var
  ix : CARDINAL;
  bit : integer;
  attc : CHAR;
begin
  ix := 1;
  SetLength(Result,0);
  while (dw <> 0) do
  begin
    bit := (dw and $00000001);
    attc := #0;
    if (bit <> 0) then
    CASE ix of
      FILE_ATTRIBUTE_ARCHIVE: attc := 'A';
      FILE_ATTRIBUTE_COMPRESSED: attc := 'C';
      FILE_ATTRIBUTE_DIRECTORY: attc := 'D';
      FILE_ATTRIBUTE_HIDDEN: attc := 'H';
      FILE_ATTRIBUTE_NORMAL: attc := 'N';
      FILE_ATTRIBUTE_OFFLINE: attc := 'O';
      FILE_ATTRIBUTE_READONLY:attc := 'R';
      FILE_ATTRIBUTE_SYSTEM: attc := 'S';
      FILE_ATTRIBUTE_TEMPORARY:attc := 'T';
    END;
    if (attc <> #0) then
      Result := Result + attc;
    dw := dw SHR 1;
    ix := ix SHL 1;
  end;
end;


procedure InitSplatProgress(var sprog : TJvSplatProgress; sn:TJvSplatNotify);
begin
  with sprog do
  begin
    filename := '';
    timeStart := 0;
    fileStart := 0;
    bytesDone := 0;
    bytesLeft := 0;
    retries  := 0;
    fileNumber := 0;
    totalFiles := 0;
    totalBytes := 0;
    totalDone := 0;
    notify := sn;
    errorlog := nil;
  end;
end;

{$IFNDEF WIN32}
function ExtractFileDrive(const path : string) : string;
begin
  if Length(path) > 0 then
    Result := path[1] + ':'
  else
    Result := '';
end;
{$ENDIF}

function GetTempNameInDirectory(const inName : string; maxnamelen:integer) : string;
{inName is the ultimate name of a file to be given an
initial unique temporary name. Assumes file will be created immediatedly
after getting name}
var
  unique : boolean;
  path : string;
  prefix : string;
  prelen : integer;
  timeNow : TDateTime;
  hour,min,sec,msec : Word;
  ct : integer;
begin
  path := ExtractFilePath(inName);
  RemoveEndChar(path,kDirSeparator);
  prefix := ExtractFileName(inName);
  if maxnamelen < 12 then
    maxnamelen := 12;
  prelen := maxnamelen - 8;
  if prelen > 4 then
    prelen := 4
  else if prelen < 0 then
    prelen := 0;
  prefix := Copy(prefix,1,prelen);
  for ct := 1 to Length(prefix) do
    if prefix[ct] = '.' then
      prefix[ct] := '_';
  while (Length(prefix) < prelen) do
    prefix := prefix + CHAR(ORD('Z')-Length(prefix));
  ct := 0;
  repeat
    timeNow := Now;
    DecodeTime(timeNow,hour,min,sec,msec);
    prefix := prefix +  IntToHex(Random(10000),4);
    prefix := prefix + IntToHex(msec,4);
    Result := Format('%s\%s.tmp', [path,Copy(prefix,1,maxnamelen-prelen-4)] );
    unique := not FileExists(Result);
    Inc(ct);
  until (unique or (ct > 1000));
  if not unique then
    raise EJvSplatError.create('Can''t make tempoary file for '+inName);
end;

function RuboutFile(const path : string) : boolean;
var
  attr :  WORD;
begin
  if FileExists(path) then
  begin { if file already exists, }
    attr := FileGetAttr(path);
    if (attr and faReadOnly) <> 0 then
    begin
      FileSetAttr(path, (attr and (not faReadOnly)));
    end;
    Result := SysUtils.DeleteFile(path);
  end
  else
    Result := TRUE;
end;

function StreamCopyProgress(Source,Dest: TStream; Count: Longint; sprog : PJvSplatProgress): Longint;
{$IFNDEF WIN32}
const
  MaxBufSize = $4000;
{$ELSE}
const
  MaxBufSize = $F000;
{$ENDIF}

var
  BufSize, N: Integer;
  Buffer: PChar;


begin
  if Count = 0 then
  begin
    Source.Position := 0;
    Count := Source.Size;
  end;


  Result := Count;
  if Count > MaxBufSize then
    BufSize := MaxBufSize
  else
    BufSize := Count;
  GetMem(Buffer, BufSize);
  try
    while Count <> 0 do
    begin
      if Count > BufSize then N := BufSize else N := Count;

      Source.ReadBuffer(Buffer^, N);

      Dest.WriteBuffer(Buffer^, N);

      Dec(Count, N);
      if sprog <> nil then
      	with sprog^ do begin
        	bytesDone := Result - Count;
            bytesLeft := Count;
        	if assigned(notify) then
            	notify(sprog);
        end;
    end;
  finally
    FreeMem(Buffer, BufSize);
  end;
end;

procedure DoSplatFile(const srcPath, destPath : string;  sprog:PJvSplatProgress);
type
  PByte = ^Byte;
var
  tempFilename : string;
  tempstr : String;
  fromName, toName,filename : string;
  toStream, fromStream : TFileStream;
  fileDate : LongInt;
  fromAttr : Byte;
  volName,sysname : PChar;
  exceptFlag : Boolean;
  copied : boolean;
  serialNo : DWORD;
  maxFileLength : DWORD;
  sysFlags : DWORD;
  retrycount : integer;
  shareMode : WORD;
  dummy : boolean;


  procedure ForceNull(var thing : TObject);
  begin
    thing := nil;
  end;

  procedure FreeStream(var iostream : TFileStream);
  var
    tempstr : TFileStream;
  begin
    tempstr := iostream;
    iostream := nil;
    tempstr.Free;
  end;

  procedure CheckException(exlog : Exception);
  begin
    Inc(retrycount);
    FreeStream(fromStream);
    FreeStream(toStream);
    if FileExists(tempFileName) then
      RuboutFile(tempFileName);
    if sprog <> nil then
    begin
      sprog^.retries := retrycount;
      if Assigned(sprog^.errorLog) then
         sprog^.errorLog(exlog.message);
    end;
  end;

begin
  if sprog <> nil then
  with sprog^ do
  begin
    retries := 0;
    filename := ExtractFileName(srcPath);
    filestart := Now;
    if timestart = 0 then
      timestart := filestart;
  end;
  tempFileName := '';
  exceptFlag := FALSE;
  TRY {for exceptions}
  {rename old file}
    tempFileName := '';
    volName := StrAlloc(MAX_PATH+1);
    sysname := StrAlloc(MAX_PATH+1);
    fromName := srcPath;
    toName := destPath;
    fromAttr := FileGetAttr(fromName);   {get source attributes}
	  tempstr := ExtractFileDrive(toName)+'\';
{$IFDEF WIN32}
    TRY
      if not GetVolumeInformation( PChar(tempstr),volName,
        MAX_PATH,
        @serialNo,
        maxFileLength,
        sysFlags,
        sysName,
        MAX_PATH) then
      begin
        raise EJvSplatError.Create('Drive ' + tempstr +' not available');
      end;
    FINALLY
      StrDispose(volName);
      StrDispose(sysname);
    END;
{$ELSE}
    maxFileLength := 12;
{$ENDIF}
    filename := ExtractFileName(toName);
    if integer(maxFileLength) < Length(filename) then
    begin
    	raise EJvSplatError.create('Filename '+toName+' is too long');
    end;
	(* always copy to temporary file first *)
    retrycount := 0;
    copied := FALSE;
    REPEAT
      tempFilename := GetTempNameInDirectory(toName,maxFileLength);
      ForceNull(TObject(toStream));
      ForceNull(TObject(fromStream));
      TRY
        TRY
          toStream := TFileStream.Create(tempFilename,fmCreate);
          shareMode := fmShareDenyWrite;
          dummy := TRUE;
          while dummy do
          begin
            TRY
              fromStream := TFileStream.Create(fromName,fmOpenRead+shareMode);
              dummy := FALSE;
              {if shareMode = fmShareDenyNone then
                MessageDlg(Format('Deny write not obtained for %s. Continue?',[fromName]),
                  mtWarning,[mbYes,mbNo],0);}
            EXCEPT
              On exopen:Exception do
                begin
                if shareMode <> fmShareDenyNone then
                  shareMode := fmShareDenyNone
                else
                  raise;
                end;
            END;
          end;
          fileDate := FileGetDate(fromStream.Handle);
        {toStream.CopyFrom(fromStream, 0);}
          StreamCopyProgress(fromStream,tostream,0,sprog);
          FileSetDate(toStream.handle, fileDate);  (* make sure has same file date *)
          copied := TRUE;
        FINALLY
          FreeStream(toStream);
          FreeStream(fromStream);
        END;
      EXCEPT
        on exopen : EFOpenError do
        begin
          CheckException(exopen);
          if (retrycount > gMaxSplatRetries) then
            raise;
        end;
        on excreate:EFCreateError do
        begin
          CheckException(excreate);
          if (retrycount > gMaxSplatRetries) then
            raise;
        end;
        on excopy:EStreamError do
        begin
          CheckException(excopy);
          if (retrycount > gMaxSplatRetries) then
            raise;
        end;
        on exall:Exception do
        begin
          CheckException(exall);
          raise;
        end;
      END;
    UNTIL (copied);
    if not RuboutFile(toName) then
      raise EJvSplatError.Create('Failed to delete original '+toName);
    if not RenameFile(tempFileName,toName) then
    begin
      (* delete attempted copy *)
      raise EJvSplatError.Create('Failed to rename '+tempFileName+' as '+toName);
    end else
    begin
      (* set attributes to be same as source *)
      FileSetAttr(toName, fromAttr);
    end;
  EXCEPT
    on ex1:Exception do
    begin
      tempstr := ex1.message;
      exceptFlag := TRUE;
    end;
  END;
  if exceptFlag then
  begin
  {try fallible stuff outside of except}
  if (Length(tempFileName) > 0) and (FileExists(tempFileName)) then
    SysUtils.DeleteFile(tempFileName);
    raise EJvSplatError.Create(tempstr); {always use own exception}
  end;
end;

procedure DoSplatDirectory(const srcPath, destPath : string;
  sprog:PJvSplatProgress; conditions : integer;
  recurse : boolean);
var
  cpa : TJvSplatDirectory;
  filter : string;
  srcFile : string;
  destFile : string;
begin
  srcFile := ExtractFileDir(srcPath);
  destFile := destPath;
  RemoveEndChar(destFile,kDirSeparator);
  if (not IsVolume(destFile)) then
  if (not DirectoryExists(destFile)) then
    begin
      ForceDirectories(destFile);
    end;
  filter := ExtractFileName(srcPath);
  cpa := TJvSplatDirectory.Create(srcFile,destFile,filter,conditions,recurse,sprog);
  try
    cpa.Execute;
  finally
    cpa.free;
  end;
end;

end.
